import asyncio
import contextlib
import os
import logging
import copy
from time import sleep, time
from typing import Any, Callable, Dict, List, Optional, Union

from anyscale.api import configure_tcp_keepalive
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute import (
    CreateClusterCompute,
)
from anyscale.sdk.anyscale_client.models.production_job import ProductionJob
from anyscale.sdk.anyscale_client.models.ha_job_states import HaJobStates
from anyscale.utils.runtime_env import upload_and_rewrite_working_dir
from anyscale.sdk import anyscale_client
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi
from anyscale.version import __version__ as version
from anyscale.sdk.anyscale_client.models.create_cluster_environment import (
    CreateClusterEnvironment,
)
from anyscale.sdk.anyscale_client.models.create_byod_cluster_environment import (
    CreateBYODClusterEnvironment,
)
from anyscale.sdk.anyscale_client.models.cluster import Cluster
from anyscale.sdk.anyscale_client.models.cluster_state import ClusterState
from anyscale.sdk.anyscale_client.models.update_cluster import UpdateCluster

from anyscale.sdk.anyscale_client.models.cluster_environment_build import (
    ClusterEnvironmentBuild,
)
from anyscale.shared_anyscale_utils.headers import RequestHeaders
from anyscale.cli_logger import BlockLogger, StringLogger, pad_string
from anyscale.cluster_compute import get_selected_cloud_id_or_default
from anyscale.util import get_endpoint
from anyscale.authenticate import AuthenticationBlock
from anyscale_client.models import ClusterEnvironmentBuildStatus
from anyscale.util import get_ray_and_py_version_for_default_cluster_env
from anyscale.project_utils import get_default_project
from anyscale.utils.ray_version_utils import get_correct_name_for_base_image_name

logger = logging.getLogger(__file__)


def _upload_and_rewrite_working_dir_in_create_production_job(create_production_job):
    """
    If a local working_dir and upload_path are specified, this method
    will upload the working_dir and rewrites the working_dir to the remote uri.

    Please note that this method mutates the model in-place.
    """
    if create_production_job.config is not None:
        runtime_env = None
        if isinstance(create_production_job.config, dict):
            runtime_env = create_production_job.config.get("runtime_env")
        elif hasattr(create_production_job.config, "runtime_env"):
            runtime_env = create_production_job.config.runtime_env
        if runtime_env is not None:
            new_runtime_env = upload_and_rewrite_working_dir(runtime_env)
            if isinstance(create_production_job.config, dict):
                create_production_job.config["runtime_env"] = new_runtime_env
            elif hasattr(create_production_job.config, "runtime_env"):
                create_production_job.config.runtime_env = new_runtime_env


def _upload_and_rewrite_working_dir_ray_serve_config(
    ray_serve_config: Optional[Dict[str, Any]]
) -> Optional[Dict[str, Any]]:
    """
    If a local working_dir and upload_path are specified, this method
    will upload the working_dir and rewrites the working_dir to the remote uri.

    Please note that this method returns the new ray_serve_config.
    """
    if not ray_serve_config:
        return ray_serve_config
    res = copy.deepcopy(ray_serve_config)
    if isinstance(res, dict):
        if "applications" in res:
            for ray_serve_app_config in res["applications"]:
                runtime_env = ray_serve_app_config.get("runtime_env")
                if runtime_env is not None:
                    new_runtime_env = upload_and_rewrite_working_dir(runtime_env)
                    ray_serve_app_config["runtime_env"] = new_runtime_env
        else:
            runtime_env = res.get("runtime_env")
            if runtime_env is not None:
                new_runtime_env = upload_and_rewrite_working_dir(runtime_env)
                res["runtime_env"] = new_runtime_env
    return res


def _is_create_byod_cluster_environment(obj: Any) -> bool:
    """
    Returns True if the object's config_json has "docker_image" set.
    The object can be either a dictionary or an object with "config_json"
    as an attribute.
    """
    config_json = None
    if isinstance(obj, dict):
        config_json = obj.get("config_json")
    else:
        config_json = getattr(obj, "config_json", None)
    return isinstance(config_json, dict) and "docker_image" in config_json


# Build states that are in progress
IN_PROGRESS_BUILD_STATES = [
    ClusterEnvironmentBuildStatus.IN_PROGRESS,
    ClusterEnvironmentBuildStatus.PENDING,
]


class AnyscaleSDK(DefaultApi):  # type: ignore
    def __init__(
        self,
        auth_token: Optional[str] = None,
        host: Optional[str] = None,
    ):
        # Default host to production but respect the ANYSCALE_HOST environment variable.
        if host is None:
            host = os.environ.get("ANYSCALE_HOST", "https://api.anyscale.com")

        # Automatically convert anyscale host to host for anyscale external api
        if host.startswith("https://console."):
            host = host.replace("console.", "api.", 1)

        # Automatically convert anyscale host to host for anyscale external api (premerge)
        if host.startswith("https://ci-"):
            host = host.rstrip("/") + "/ext"

        # Adds base path "v0" for API endpoints
        endpoint_url = host.rstrip("/") + "/v0"
        configuration = anyscale_client.Configuration(host=endpoint_url)
        configuration.proxy = os.environ.get("https_proxy")
        configuration.connection_pool_maxsize = 100
        if auth_token is None:
            auth_token, _ = AuthenticationBlock._load_credentials()
        api_client = anyscale_client.ApiClient(
            configuration, cookie=f"cli_token={auth_token}"
        )
        configure_tcp_keepalive(api_client)
        api_client.set_default_header(RequestHeaders.CLIENT, "SDK")
        api_client.set_default_header(RequestHeaders.CLIENT_VERSION, version)

        self.log: BlockLogger = BlockLogger()
        self._auth_token = auth_token

        super(AnyscaleSDK, self).__init__(api_client)

    def build_cluster_environment(
        self,
        create_cluster_environment: Union[
            CreateClusterEnvironment, CreateBYODClusterEnvironment
        ],
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        log_output: bool = False,
    ) -> Optional[ClusterEnvironmentBuild]:
        """
        Creates a new Cluster Environment and waits for build to complete.

        If a Cluster Environment with the same name already exists, this will
        create an updated build of that environment.

        Args:
            create_cluster_environment - CreateClusterEnvironment object
            poll_rate_seconds - seconds to wait when polling build operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for build operation to complete before timing out; defaults to no timeout

        Returns:
            Newly created ClusterEnvironmentBuild object

        Raises:
            Exception if building Cluster Environment failed or timed out
        """

        self.log.log_output = log_output
        cluster_environments = self.search_cluster_environments(
            {
                "name": {"equals": create_cluster_environment.name},
                "paging": {"count": 1},
            }
        ).results

        if not cluster_environments:
            self.log.info(
                f"Creating new cluster environment {create_cluster_environment.name}"
            )
            if _is_create_byod_cluster_environment(create_cluster_environment):
                cluster_environment = self.create_byod_cluster_environment(
                    create_cluster_environment
                ).result
            else:
                base_image_name = create_cluster_environment.config_json['base_image']
                correct_base_image_name = get_correct_name_for_base_image_name(base_image_name)
                if base_image_name != correct_base_image_name: # We need to make sure base_image_name after ray 2.7.0, before 2.8.0 have a optimized suffix
                    self.log.error(f"Starting from Ray version 2.7.0 and before version 2.8.0, you might be required to include the suffix 'optimized' in the base_image name for that specific Ray version. Try using {self.log.highlight(correct_base_image_name)}")
                    return  # type: ignore
                cluster_environment = self.create_cluster_environment(
                    create_cluster_environment
                ).result
            build = self.list_cluster_environment_builds(
                cluster_environment.id
            ).results[0]
            build_operation_id = build.id
        else:
            self.log.info(
                f"Building new revision of cluster environment {create_cluster_environment.name}"
            )
            cluster_environment = cluster_environments[0]
            if _is_create_byod_cluster_environment(create_cluster_environment):
                build = self.create_byod_cluster_environment_build(
                    {
                        "cluster_environment_id": cluster_environment.id,
                        "config_json": create_cluster_environment.config_json,
                    }
                ).result
            else:
                base_image_name = create_cluster_environment.config_json['base_image']
                correct_base_image_name = get_correct_name_for_base_image_name(base_image_name)
                if base_image_name != correct_base_image_name: # We need to make sure base_image_name after ray 2.7.0, before 2.8.0 have a optimized suffix
                    self.log.error(f"Starting from Ray version 2.7.0 and before version 2.8.0, you might be required to include the suffix 'optimized' in the base_image name for that specific Ray version. Try using {self.log.highlight(correct_base_image_name)}")
                    return  # type: ignore
                build = self.create_cluster_environment_build(
                    {
                        "cluster_environment_id": cluster_environment.id,
                        "config_json": create_cluster_environment.config_json,
                    }
                ).result
            build_operation_id = build.id

        build = self._wait_for_cluster_environment_build_operation(
            build_operation_id, poll_rate_seconds, timeout_seconds, log_output
        )
        self.log.close_block()

        block_label = "BuildDetails"
        self.log.open_block(block_label=block_label, block_title="Build Details")
        with self.log.indent():
            pad_len = len("Name:")
            self.log.info(
                pad_string("Name:", pad_len),
                BlockLogger.highlight(f"{create_cluster_environment.name}:{build.revision}"),
                block_label=block_label
            )
            self.log.info(
                pad_string("ID:", pad_len),
                BlockLogger.highlight(build.id),
                block_label=block_label
            )
        return build

    def _wait_for_cluster_environment_build_operation(
        self,
        operation_id: str,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        log_output: bool = False,
    ) -> ClusterEnvironmentBuild:
        """
        Waits for a Cluster Environment Build operation to complete.

        Args:
            operation_id - ID of the Cluster Environment Build operation
            poll_rate_seconds - seconds to wait when polling build operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for build operation to complete before timing out; defaults to no timeout

        Returns:
            ClusterEnvironmentBuild object

        Raises:
            Exception if building Cluster Environment fails or times out.
        """

        self.log.log_output = log_output
        timeout = time() + timeout_seconds if timeout_seconds else None

        operation = self.get_build(operation_id)
        url = get_endpoint(f"configurations/app-config-details/{operation_id}")
        self.log.info(
            f"Waiting for cluster environment to build. View progress at {url}"
        )
        self.log.info(f"Status: {operation.result.status}")
        while operation.result.status in IN_PROGRESS_BUILD_STATES:
            if timeout and time() > timeout:
                raise Exception(
                    f"Building Cluster Environment timed out after {timeout_seconds} seconds."
                )

            sleep(poll_rate_seconds)
            self.log.info(f"Status: {operation.result.status}")
            operation = self.get_build(operation_id)
        if operation.result.status != ClusterEnvironmentBuildStatus.SUCCEEDED:
            raise Exception(
                f"Failed to build Cluster Environment, you can check the full logs at: {url}."
            )
        else:
            self.log.info("Cluster environment successfully finished building.")
            return self.get_build(operation_id).result

    def launch_cluster(
        self,
        project_id: Optional[str],
        cluster_name: str,
        cluster_environment_build_id: Optional[str] = None,
        cluster_compute_id: Optional[str] = None,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        idle_timeout_minutes: Optional[int] = None,
        cluster_compute_config: Optional[ClusterComputeConfig] = None,
        log_output: bool = False,
    ) -> Cluster:
        """
        Starts a Cluster in the specified Project.
        If a Cluster with the specified name already exists, we will update that Cluster.
        Otherwise, a new Cluster will be created.

        Args:
            project_id - ID of the Project the Cluster belongs to or None to launch cluster without a project.
            cluster_name - Name of the Cluster
            cluster_environment_build_id - Cluster Environment Build to start this Cluster with
                                           If none, uses a default cluster_environment_build
            cluster_compute_id - Cluster Compute to start this Cluster with
                                 If none, it checks if `cluster_compute_config` is specified.
                                 If `cluster_compute_config` is none, it uses a default cluster_compute.
            poll_rate_seconds - seconds to wait when polling Cluster operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for Cluster operation to complete before timing out; defaults to no timeout
            idle_timeout_minutes - Idle timeout (in minutes), after which the Cluster is terminated
            cluster_compute_config - One-off Cluster Compute that this Cluster will use.
            log_output - Whether to show verbose logs during execution

        Returns:
            Cluster object

        Raises:
            Exception if starting Cluster fails or times out
        """

        self.log.log_output = log_output
        if project_id:
            search_project_id = project_id
        else:
            parent_cloud_id = get_selected_cloud_id_or_default(
                cluster_compute_id=cluster_compute_id,
                cluster_compute_config=cluster_compute_config,
                cloud_id=None,
                cloud_name=None,
            )
            default_project = get_default_project(parent_cloud_id=parent_cloud_id)
            project_id = default_project.id

        clusters: Optional[List[Cluster]] = self.search_clusters(
            {"project_id": project_id, "name": {"equals": cluster_name}}
        ).results

        if clusters:

            sorted_clusters_list = sorted(clusters, key=lambda x: x.created_at, reverse=True)
            cluster: Cluster = sorted_clusters_list[0]

            if cluster.state in [ClusterState.STARTINGUP, ClusterState.RUNNING, ClusterState.UPDATING, ClusterState.TERMINATING, ClusterState.AWAITINGSTARTUP]:
                raise Exception(
                    f"Cluster {cluster_name} in project {project_id} is in state {cluster.state}. "
                    "Please terminate the cluster or use a new cluster name, then rerun launch_cluster."
                )

            if not cluster_environment_build_id:
                cluster_environment_build_id = cluster.cluster_environment_build_id

            if not cluster_compute_id:
                cluster_compute_id = cluster.cluster_compute_id

            if not idle_timeout_minutes:
                idle_timeout_minutes = cluster.idle_timeout_minutes

            # Update cluster with new configurations
            # Or use configurations from terminated cluster
            update_cluster_config = UpdateCluster(
                cluster_environment_build_id=cluster_environment_build_id,
                cluster_compute_id=cluster_compute_id,
                idle_timeout_minutes=idle_timeout_minutes,
            )

            self.update_cluster(cluster.id, update_cluster_config)

        else:
            if not cluster_environment_build_id:
                # Use default cluster environment build when starting cluster without specifying an id
                (
                    ray_version,
                    py_version,
                ) = get_ray_and_py_version_for_default_cluster_env()
                cluster_environment_build_id = (
                    self.get_default_cluster_environment_build(
                        f"py{py_version}", ray_version
                    ).result.id
                )

            if not cluster_compute_id:
                if cluster_compute_config:
                    cluster_compute_id = self.create_cluster_compute(
                        CreateClusterCompute(
                            config=cluster_compute_config, anonymous=True
                        )
                    ).result.id
                else:
                    # Use default cluster compute when starting cluster without specifying an id
                    cluster_compute_id = self.get_default_cluster_compute().result.id


            create_cluster_payload: Dict[str, Any] = {
                "name": cluster_name,
                "project_id": project_id,
                "cluster_environment_build_id": cluster_environment_build_id,
                "cluster_compute_id": cluster_compute_id,
            }
            if idle_timeout_minutes:
                create_cluster_payload["idle_timeout_minutes"] = idle_timeout_minutes

            # Create cluster DB reference
            cluster = self.create_cluster(create_cluster_payload).result

        # Start cluster
        start_operation = self.start_cluster(
            cluster.id,
            {
                "cluster_environment_build_id": cluster_environment_build_id,
                "cluster_compute_id": cluster_compute_id,
                "idle_timeout_minutes": idle_timeout_minutes,
            },
        ).result

        return self.wait_for_cluster_operation(
            start_operation.id, poll_rate_seconds, timeout_seconds
        )

    def launch_cluster_with_new_cluster_environment(
        self,
        project_id: Optional[str],
        cluster_name: str,
        create_cluster_environment: CreateClusterEnvironment,
        cluster_compute_id: Optional[str] = None,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        idle_timeout_minutes: Optional[int] = None,
        cluster_compute_config: Optional[ClusterComputeConfig] = None,
    ) -> Cluster:
        """
        Builds a new Cluster Environment, then starts a Cluster in the specified Project with the new build.
        If a Cluster with the specified name already exists, we will update that Cluster.
        Otherwise, a new Cluster will be created.

        Args:
            project_id - ID of the Project the Cluster belongs to or None to launch a cluster without a project.
            cluster_name - Name of the Cluster
            create_cluster_environment - CreateClusterEnvironment object
            cluster_compute_id - Cluster Compute to start this Cluster with
                                 If none, it checks if `cluster_compute_config` is specified.
                                 If `cluster_compute_config` is  none, it uses a default cluster_compute.
            poll_rate_seconds - seconds to wait when polling for operations; defaults to 15
            timeout_seconds - maximum number of seconds to wait for each operation to complete before timing out; defaults to no timeout
            idle_timeout_minutes - Idle timeout (in minutes), after which the Cluster is terminated
            cluster_compute_config - One-off Cluster Compute that this Cluster will use.

        Returns:
            Cluster object

        Raises:
            Exception if building the new Cluster Environment fails or starting the Cluster fails.
        """

        cluster_environment_build = self.build_cluster_environment(
            create_cluster_environment, poll_rate_seconds, timeout_seconds
        )

        assert cluster_environment_build is not None
        return self.launch_cluster(
            project_id,
            cluster_name,
            cluster_environment_build.id,
            cluster_compute_id,
            poll_rate_seconds,
            timeout_seconds,
            idle_timeout_minutes,
            cluster_compute_config,
        )

    def wait_for_cluster_operation(
        self,
        operation_id: str,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
    ) -> Cluster:
        """
        Waits for a Cluster operation to complete, most commonly used when starting, terminating, or updating a Cluster.

        Args:
            operation_id - ID of the Cluster Operation
            poll_rate_seconds - seconds to wait when polling build operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for build operation to complete before timing out; defaults to no timeout

        Returns:
            Cluster object when the operation completes successfully

        Raises:
            Exception if building Cluster operation fails or times out
        """

        timeout = time() + timeout_seconds if timeout_seconds else None

        operation = self.get_cluster_operation(operation_id).result
        while not operation.completed:
            if timeout and time() > timeout:
                raise Exception(
                    f"Cluster start up timed out after {timeout_seconds} seconds."
                )

            sleep(poll_rate_seconds)
            operation = self.get_cluster_operation(operation_id).result

        if operation.result.error:
            raise Exception("Failed to start Cluster", operation.result.error)
        else:
            return self.get_cluster(operation.cluster_id).result

    def fetch_production_job_logs(self, job_id: str) -> str:
        from anyscale.controllers.job_controller import JobController

        """
        Retrieves logs for a job (eg. 'prodjob_123').
        - This will retrieve logs from only the latest job run of the job.

        Args:
            job_id (str): ID of a job

        Returns:
            str: All logs for the job
                (so far, if the job is still running)
        """
        string_logger = StringLogger()
        job_controller = JobController(string_logger, auth_token=self._auth_token)
        job_controller.logs(job_id, no_cluster_journal_events=True)

        return string_logger.out_string

    def create_job(self, create_production_job, **kwargs):
        _upload_and_rewrite_working_dir_in_create_production_job(create_production_job)
        return super().create_job(create_production_job, **kwargs)

    def create_service(self, create_production_service, **kwargs):
        _upload_and_rewrite_working_dir_in_create_production_job(
            create_production_service
        )
        return super().create_service(create_production_service, **kwargs)

    def apply_service(self, create_production_service, **kwargs):
        _upload_and_rewrite_working_dir_in_create_production_job(
            create_production_service
        )
        return super().apply_service(create_production_service, **kwargs)

    def rollout_service_v2(self, apply_production_service_v2_model, **kwargs):
        """
        DEPRECATED. Please use rollout_service.
        """
        if isinstance(apply_production_service_v2_model, dict):
            ray_serve_config = apply_production_service_v2_model.get("ray_serve_config")
            new_ray_serve_config = _upload_and_rewrite_working_dir_ray_serve_config(
                ray_serve_config
            )
            apply_production_service_v2_model["ray_serve_config"] = new_ray_serve_config
        elif hasattr(apply_production_service_v2_model, "ray_serve_config"):
            ray_serve_config = apply_production_service_v2_model.ray_serve_config
            new_ray_serve_config = _upload_and_rewrite_working_dir_ray_serve_config(
                ray_serve_config
            )
            apply_production_service_v2_model.ray_serve_config = new_ray_serve_config

        return super().rollout_service_v2(apply_production_service_v2_model, **kwargs)

    def rollout_service(self, apply_service_model, **kwargs):
        """
        If a local working_dir and upload_path are specified in ray_serve_config,
        this method uploads the local working_dir to the upload_path and rewrites
        the working_dir in ray_serve_config.
        """
        if isinstance(apply_service_model, dict):
            ray_serve_config = apply_service_model.get("ray_serve_config")
            new_ray_serve_config = _upload_and_rewrite_working_dir_ray_serve_config(
                ray_serve_config
            )
            apply_service_model["ray_serve_config"] = new_ray_serve_config
        elif hasattr(apply_service_model, "ray_serve_config"):
            ray_serve_config = apply_service_model.ray_serve_config
            new_ray_serve_config = _upload_and_rewrite_working_dir_ray_serve_config(
                ray_serve_config
            )
            apply_service_model.ray_serve_config = new_ray_serve_config

        return super().rollout_service(apply_service_model, **kwargs)

    def get_cluster_token(self, project_id: str, cluster_name: str) -> str:
        """
        Retrieves cluster token for a Cluster.
        This cluster token can be used to authenticate Anyscale services (eg: serve, jupyter, grafana)

        Args:
            project_id - ID of the Project the Cluster belongs to
            cluster_name - Name of the Cluster

        Returns
            string containing cluster token

        Raises
            Exception if invalid cluster name
        """
        clusters = self.search_clusters(
            {"project_id": project_id, "name": {"equals": cluster_name}}
        ).results

        if clusters:
            cluster = clusters[0]
        else:
            raise Exception(
                (
                    f"Cluster named {cluster_name} does not exist in Project {project_id} "
                    "or this user does not have read permissions for it."
                )
            )
        cluster_token = cluster.access_token
        return cluster_token



from typing import Dict, List, Optional

from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk import sdk_command
from anyscale.job_queue._private.job_queue_sdk import PrivateJobQueueSDK
from anyscale.job_queue.models import JobQueueStatus


_JOB_QUEUE_SDK_SINGLETON_KEY = "job_queue_sdk"


_LIST_EXAMPLE = """
import anyscale

# Example: List the first 50 job queues
for jq in anyscale.job_queue.list(max_items=50):
    print(jq.id, jq.name, jq.state)
"""

_LIST_ARG_DOCSTRINGS = {
    "job_queue_id": "If provided, fetches only the job queue with this ID.",
    "name": "Filter by job queue name.",
    "creator_id": "Filter by the user ID of the creator.",
    "cloud": "Filter by cloud name.",
    "project": "Filter by project name.",
    "cluster_status": "Filter by the state of the associated cluster.",
    "tags_filter": "Filter by tags. Accepts dict[key] -> List[values] or list['key:value'] entries.",
    "page_size": "Number of items per API request page.",
    "max_items": "Maximum total number of items to return.",
    "sorting_directives": "List of directives to sort the results.",
}

_STATUS_EXAMPLE = """
import anyscale

status = anyscale.job_queue.status(job_queue_id=\"jobq_abc123\")
print(status)
"""

_STATUS_ARG_DOCSTRINGS = {
    "job_queue_id": "The unique ID of the job queue.",
}

_UPDATE_EXAMPLE = """
import anyscale

updated_jq = anyscale.job_queue.update(job_queue_id=\"jobq_abc123\", max_concurrency=5)
print(updated_jq)
"""

_UPDATE_ARG_DOCSTRINGS = {
    "job_queue_id": "ID of the job queue to update.",
    "job_queue_name": "Name of the job queue to update (alternative to ID).",
    "max_concurrency": "New maximum concurrency value.",
    "idle_timeout_s": "New idle timeout in seconds.",
}

_TAGS_ADD_EXAMPLE = """
import anyscale

anyscale.job_queue.add_tags(job_queue_id="jobq_abc123", tags={"team": "mlops"})
"""

_TAGS_ADD_ARG_DOCSTRINGS = {
    "job_queue_id": "ID of the job queue to tag (alternative to name).",
    "name": "Name of the job queue to tag (alternative to ID).",
    "tags": "Key/value tags to upsert as a map {key: value}.",
}

_TAGS_REMOVE_EXAMPLE = """
import anyscale

anyscale.job_queue.remove_tags(job_queue_id="jobq_abc123", keys=["team"])
"""

_TAGS_REMOVE_ARG_DOCSTRINGS = {
    "job_queue_id": "ID of the job queue to modify (alternative to name).",
    "name": "Name of the job queue to modify (alternative to ID).",
    "keys": "List of tag keys to remove.",
}


@sdk_command(
    _JOB_QUEUE_SDK_SINGLETON_KEY,
    PrivateJobQueueSDK,
    doc_py_example=_LIST_EXAMPLE,
    arg_docstrings=_LIST_ARG_DOCSTRINGS,
)
def list(  # noqa: A001
    *,
    job_queue_id: Optional[str] = None,
    name: Optional[str] = None,
    creator_id: Optional[str] = None,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    cluster_status: Optional[str] = None,
    tags_filter: Optional[Dict[str, List[str]]] = None,
    page_size: Optional[int] = None,
    max_items: Optional[int] = None,
    sorting_directives: Optional[List[str]] = None,
    _private_sdk: Optional[PrivateJobQueueSDK] = None,
) -> ResultIterator[JobQueueStatus]:
    """List job queues or fetch a single job queue by ID."""
    return _private_sdk.list(  # type: ignore
        job_queue_id=job_queue_id,
        name=name,
        creator_id=creator_id,
        cloud=cloud,
        project=project,
        cluster_status=cluster_status,
        tags_filter=tags_filter,
        page_size=page_size,
        max_items=max_items,
        sorting_directives=sorting_directives,
    )


@sdk_command(
    _JOB_QUEUE_SDK_SINGLETON_KEY,
    PrivateJobQueueSDK,
    doc_py_example=_STATUS_EXAMPLE,
    arg_docstrings=_STATUS_ARG_DOCSTRINGS,
)
def status(
    job_queue_id: str, _private_sdk: Optional[PrivateJobQueueSDK] = None
) -> JobQueueStatus:
    """Get the status and details for a specific job queue."""
    return _private_sdk.status(  # type: ignore
        job_queue_id=job_queue_id
    )


@sdk_command(
    _JOB_QUEUE_SDK_SINGLETON_KEY,
    PrivateJobQueueSDK,
    doc_py_example=_UPDATE_EXAMPLE,
    arg_docstrings=_UPDATE_ARG_DOCSTRINGS,
)
def update(
    *,
    job_queue_id: Optional[str] = None,
    job_queue_name: Optional[str] = None,
    max_concurrency: Optional[int] = None,
    idle_timeout_s: Optional[int] = None,
    _private_sdk: Optional[PrivateJobQueueSDK] = None,
) -> JobQueueStatus:
    """Update a job queue."""
    return _private_sdk.update(  # type: ignore
        job_queue_id=job_queue_id,
        job_queue_name=job_queue_name,
        max_concurrency=max_concurrency,
        idle_timeout_s=idle_timeout_s,
    )


@sdk_command(
    _JOB_QUEUE_SDK_SINGLETON_KEY,
    PrivateJobQueueSDK,
    doc_py_example=_TAGS_ADD_EXAMPLE,
    arg_docstrings=_TAGS_ADD_ARG_DOCSTRINGS,
)
def add_tags(
    *,
    job_queue_id: Optional[str] = None,
    name: Optional[str] = None,
    tags: Dict[str, str],
    _private_sdk: Optional[PrivateJobQueueSDK] = None,
):
    """Upsert tags for a job queue."""
    return _private_sdk.add_tags(job_queue_id=job_queue_id, name=name, tags=tags)  # type: ignore


@sdk_command(
    _JOB_QUEUE_SDK_SINGLETON_KEY,
    PrivateJobQueueSDK,
    doc_py_example=_TAGS_REMOVE_EXAMPLE,
    arg_docstrings=_TAGS_REMOVE_ARG_DOCSTRINGS,
)
def remove_tags(
    *,
    job_queue_id: Optional[str] = None,
    name: Optional[str] = None,
    keys: List[str],
    _private_sdk: Optional[PrivateJobQueueSDK] = None,
):
    """Remove tags by key from a job queue."""
    return _private_sdk.remove_tags(job_queue_id=job_queue_id, name=name, keys=keys)  # type: ignore


_TAGS_LIST_EXAMPLE = """
import anyscale

tags: dict[str, str] = anyscale.job_queue.list_tags(name="my-queue")
"""

_TAGS_LIST_ARG_DOCSTRINGS = {
    "job_queue_id": "ID of the job queue to read tags (alternative to name).",
    "name": "Name of the job queue to read tags (alternative to ID).",
}


@sdk_command(
    _JOB_QUEUE_SDK_SINGLETON_KEY,
    PrivateJobQueueSDK,
    doc_py_example=_TAGS_LIST_EXAMPLE,
    arg_docstrings=_TAGS_LIST_ARG_DOCSTRINGS,
)
def list_tags(
    *,
    job_queue_id: Optional[str] = None,
    name: Optional[str] = None,
    _private_sdk: Optional[PrivateJobQueueSDK] = None,
) -> Dict[str, str]:
    """List tags for a job queue as a key/value mapping."""
    return _private_sdk.list_tags(job_queue_id=job_queue_id, name=name)  # type: ignore



"""
CLI commands for SCIM-related operations.

These commands handle SCIM-specific functionality like enforcing
user group permissions after SCIM is enabled.
"""

import click

from anyscale._private.anyscale_client import AnyscaleClient
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand


log = BlockLogger()


@click.group(
    "scim", help="Manage SCIM (System for Cross-domain Identity Management) settings."
)
def scim_cli() -> None:
    pass


@scim_cli.command(
    name="enforce-groups",
    cls=AnyscaleCommand,
    example=command_examples.SCIM_ENFORCE_GROUP_PERMISSIONS_EXAMPLE,
    is_beta=True,
)
def enforce_group_permissions() -> None:
    """
    Enforce SCIM-based user group permissions by removing individual user permissions.

    This command removes ALL direct user permissions so that users only derive
    permissions from their user groups.
    """
    client = AnyscaleClient()

    try:
        click.echo(
            "\n"
            "╭─────────────────── ⚠️  Confirmation Required ───────────────────╮\n"
            "│ WARNING: This is a destructive operation that cannot be undone. │\n"
            "│                                                                 │\n"
            "│ All role bindings on users will be removed.                     │\n"
            "│ Role bindings on user groups and service accounts are unchanged.│\n"
            "╰─────────────────────────────────────────────────────────────────╯\n"
        )
        click.confirm(
            "Do you want to proceed?", default=False, abort=True,
        )

        log.info("Starting SCIM permission migration...")
        response = client.migrate_scim_permissions(dry_run=False)

        result = response.get("result", response)
        errors = result.get("errors", [])

        if errors:
            raise click.ClickException("; ".join(errors))

        log.info("SCIM permission migration completed successfully.")

    except (click.ClickException, click.Abort):
        raise
    except Exception as e:  # noqa: BLE001
        log.error(f"Failed to migrate SCIM permissions: {e}")
        raise click.ClickException(str(e))



import concurrent.futures
from functools import partial
import os
import subprocess
import tempfile
from types import SimpleNamespace
from typing import cast, ClassVar, Dict, List, Optional, Tuple, Union

import click

from anyscale._private.models.model_base import ResultIterator
from anyscale._private.workload.workload_sdk import WorkloadSDK
from anyscale.client.openapi_client.models import (
    CreateExperimentalWorkspace,
    ExperimentalWorkspace,
)
from anyscale.client.openapi_client.models.resource_tag_resource_type import (
    ResourceTagResourceType,
)
from anyscale.client.openapi_client.models.session_state import SessionState
from anyscale.utils.runtime_env import parse_requirements_file
from anyscale.workspace.models import (
    UpdateWorkspaceConfig,
    Workspace,
    WorkspaceConfig,
    WorkspaceSortField,
    WorkspaceSortOrder,
    WorkspaceState,
)


SSH_TEMPLATE = """
Host Head-Node
  HostName {head_node_ip}
  User ubuntu
  IdentityFile {key_path}
  StrictHostKeyChecking false
  UserKnownHostsFile /dev/null
  IdentitiesOnly yes
  LogLevel ERROR

Host {name}
  HostName 0.0.0.0
  ProxyJump Head-Node
  Port 5020
  User ray
  IdentityFile {key_path}
  StrictHostKeyChecking false
  UserKnownHostsFile /dev/null
  IdentitiesOnly yes
  LogLevel ERROR
"""

# TODO(bryce): These options are already stated in the SSH_TEMPLATE, but we
# still pass them in the ssh command. It should be safe to remove them.
ANYSCALE_WORKSPACES_SSH_OPTIONS = [
    "-o",
    "StrictHostKeyChecking=no",
    "-o",
    "UserKnownHostsFile=/dev/null",
    "-o",
    "IdentitiesOnly=yes",
]


class PrivateWorkspaceSDK(WorkloadSDK):
    _POLLING_INTERVAL_SECONDS = 10.0
    _WAIT_TIMEOUT_SECONDS = 1800.0

    _BACKEND_SESSION_STATE_TO_WORKSPACE_STATE: ClassVar[
        Dict[SessionState, WorkspaceState]
    ] = {
        SessionState.STOPPED: WorkspaceState.TERMINATED,
        SessionState.TERMINATED: WorkspaceState.TERMINATED,
        SessionState.STARTINGUP: WorkspaceState.STARTING,
        SessionState.STARTUPERRORED: WorkspaceState.ERRORED,
        SessionState.RUNNING: WorkspaceState.RUNNING,
        SessionState.UPDATING: WorkspaceState.UPDATING,
        SessionState.UPDATINGERRORED: WorkspaceState.ERRORED,
        SessionState.STOPPING: WorkspaceState.TERMINATING,
        SessionState.TERMINATING: WorkspaceState.TERMINATING,
        SessionState.AWAITINGSTARTUP: WorkspaceState.STARTING,
        SessionState.AWAITINGFILEMOUNTS: WorkspaceState.STARTING,
        SessionState.TERMINATINGERRORED: WorkspaceState.ERRORED,
        SessionState.STOPPINGERRORED: WorkspaceState.ERRORED,
    }

    def _resolve_to_workspace_model(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> ExperimentalWorkspace:
        if name is None and id is None:
            raise ValueError("One of 'name' or 'id' must be provided.")

        if name is not None and id is not None:
            raise ValueError("Only one of 'name' or 'id' can be provided.")

        if id is not None and (cloud is not None or project is not None):
            raise ValueError("'cloud' and 'project' should only be used with 'name'.")

        model: Optional[ExperimentalWorkspace] = self.client.get_workspace(
            name=name, id=id, cloud=cloud, project=project
        )
        if model is None:
            if name is not None:
                raise ValueError(f"Workspace with name '{name}' was not found.")
            else:
                raise ValueError(f"Workspace with ID '{id}' was not found.")

        return model

    def create(self, config: WorkspaceConfig) -> str:
        if not config or config.name is None:
            raise ValueError("Workspace name must be configured")

        name = config.name

        compute_config_id, cloud_id = self.resolve_compute_config_and_cloud_id(
            compute_config=config.compute_config, cloud=config.cloud,  # type: ignore
        )

        project_id = self.client.get_project_id(
            parent_cloud_id=cloud_id, name=config.project
        )

        build_id = None

        if config.containerfile is not None:
            build_id = self._image_sdk.build_image_from_containerfile(
                name=f"image-for-workspace-{name}",
                containerfile=self.get_containerfile_contents(config.containerfile),
                ray_version=config.ray_version,
            )
        elif config.image_uri is not None:
            build_id = self._image_sdk.registery_image(
                image_uri=config.image_uri,
                registry_login_secret=config.registry_login_secret,
                ray_version=config.ray_version,
            )

        dynamic_requirements = None
        if (
            config.requirements
            and self._image_sdk.enable_image_build_for_tracked_requirements
        ):
            requirements = (
                parse_requirements_file(config.requirements)
                if isinstance(config.requirements, str)
                else config.requirements
            )
            if requirements:
                build_id = self._image_sdk.build_image_from_requirements(
                    name=f"image-for-workspace-{name}",
                    base_build_id=self.client.get_default_build_id(),
                    requirements=requirements,
                )
        elif config.requirements:
            dynamic_requirements = (
                parse_requirements_file(config.requirements)
                if isinstance(config.requirements, str)
                else config.requirements
            )

        if build_id is None:
            build_id = self.client.get_default_build_id()

        workspace_id = self.client.create_workspace(
            model=CreateExperimentalWorkspace(
                name=name,
                project_id=project_id,
                compute_config_id=compute_config_id,
                cluster_environment_build_id=build_id,
                idle_timeout_minutes=config.idle_termination_minutes,
                cloud_id=cloud_id,
                tags=config.tags,
                skip_start=True,
            )
        )

        self._logger.info(f"Workspace created successfully id: {workspace_id}")

        if dynamic_requirements:
            self.client.update_workspace_dependencies_offline_only(
                workspace_id=workspace_id, requirements=dynamic_requirements
            )
            self._logger.info(f"Applied dynamic requirements to workspace id: {name}")
        if config.env_vars:
            self.client.update_workspace_env_vars_offline_only(
                workspace_id=workspace_id, env_vars=config.env_vars
            )
            self._logger.info(f"Applied environment variables to workspace id: {name}")

        return workspace_id

    def start(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        self.client.start_workspace(workspace_model.id)
        self.logger.info(f"Starting workspace '{workspace_model.name}'")
        return workspace_model.id

    def terminate(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        self.client.terminate_workspace(workspace_model.id)
        self.logger.info(f"Terminating workspace '{workspace_model.name}'")
        return workspace_model.id

    def status(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )

        status = self._get_workspace_status(workspace_model.id)
        return status

    def wait(  # noqa: PLR0912
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        timeout_s: float = _WAIT_TIMEOUT_SECONDS,
        state: Union[str, WorkspaceState] = WorkspaceState.RUNNING,
        interval_s: float = _POLLING_INTERVAL_SECONDS,
    ):
        if not isinstance(timeout_s, (int, float)):
            raise TypeError("timeout_s must be a float")
        if timeout_s <= 0:
            raise ValueError("timeout_s must be >= 0")

        if not isinstance(interval_s, (int, float)):
            raise TypeError("interval_s must be a float")
        if interval_s <= 0:
            raise ValueError("interval_s must be >= 0")

        if isinstance(state, str):
            try:
                state = WorkspaceState.validate(state)
            except KeyError:
                raise ValueError(f"Invalid state: {state}")

        if not isinstance(state, WorkspaceState):
            raise TypeError("'state' must be a WorkspaceState.")

        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )

        workspace_id_or_name = workspace_model.id or workspace_model.name
        curr_state = self._get_workspace_status(workspace_model.id)

        self.logger.info(
            f"Waiting for workspace '{workspace_id_or_name}' to reach target state {state}, currently in state: {curr_state}"
        )
        for _ in self.timer.poll(timeout_s=timeout_s, interval_s=interval_s):
            new_state = self._get_workspace_status(workspace_model.id)

            if new_state != curr_state:
                self.logger.info(
                    f"Workspace '{workspace_id_or_name}' transitioned from {curr_state} to {new_state}"
                )
                curr_state = new_state

            if curr_state == state:
                self.logger.info(
                    f"Workspace '{workspace_id_or_name}' reached target state, exiting"
                )
                break
        else:
            raise TimeoutError(
                f"Workspace '{workspace_id_or_name}' did not reach target state {state} within {timeout_s}s. Last seen state: {curr_state}."
            )

    def _get_workspace_status(self, workspace_id: Optional[str]) -> WorkspaceState:

        cluster = self.client.get_workspace_cluster(workspace_id)
        if not cluster:
            raise ValueError(f"Workspace {workspace_id} cluster not found.")
        return self._convert_cluster_state_to_workspace_state(cluster.state)

    def generate_ssh_config_file(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        ssh_config_path: Optional[str] = None,
    ) -> Tuple[str, str]:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )

        assert (
            self.status(id=workspace_model.id) == WorkspaceState.RUNNING
        ), "Workspace must be running to generate SSH config file"

        head_node_ip = self.client.get_cluster_head_node_ip(workspace_model.cluster_id)
        ssh_key = self.client.get_cluster_ssh_key(workspace_model.cluster_id)

        def _store_ssh_key(name: str, ssh_key: str, key_dir: str) -> str:
            key_path = os.path.join(key_dir, f"{name}.pem")
            os.makedirs(os.path.dirname(key_path), exist_ok=True)

            with open(key_path, "w", opener=partial(os.open, mode=0o600)) as f:
                f.write(ssh_key)

            return key_path

        if ssh_config_path is None:
            ssh_config_path = tempfile.mkdtemp()

        key_path = _store_ssh_key(
            ssh_key.key_name, ssh_key.private_key, ssh_config_path
        )

        ssh_config = SSH_TEMPLATE.format(
            head_node_ip=head_node_ip, key_path=key_path, name=workspace_model.name
        )

        config_file_name = os.path.join(ssh_config_path, "config")
        with open(config_file_name, "w") as f:
            f.write(ssh_config)

        return workspace_model.name, config_file_name

    def run_command(
        self,
        command: str,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        host_name, config_file = self.generate_ssh_config_file(id=workspace_model.id)
        return subprocess.run(
            ["ssh"]
            + ANYSCALE_WORKSPACES_SSH_OPTIONS
            + ["-F", config_file, host_name, command],
            check=kwargs.pop("check", False),
            **kwargs,
        )

    def get_default_dir_name(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        return self.client.get_workspace_default_dir_name(workspace_model.id)

    def _parse_rsync_dry_run_output(
        self, dry_run_output: str
    ) -> Tuple[List[str], List[str]]:
        """Parse rsync dry-run output to detect file changes vs additions.
        Note that --itemize-changes is needed to detect file changes.
        Format is like "<fcsT...... test.py" where:
        - First char indicates operations (>, <, *)
        - Third char (c or .) indicates if checksum changed (only if --checksum is used)

        Returns:
            Tuple of (modifying_files, deleting_files)
        """
        modifying_files = []
        deleting_files = []

        ITEMISZED_PREFIX_LEN = 12

        changes = dry_run_output.strip().split("\n")
        for line in changes:
            if (
                not line or len(line) < ITEMISZED_PREFIX_LEN
            ):  # Need at least "<fcsT......" part
                continue
            if line.startswith("*deleting"):
                deleting_files.append(line[ITEMISZED_PREFIX_LEN:])
            elif line[2] == "c":  # Check the checksum position
                modifying_files.append(line[ITEMISZED_PREFIX_LEN:])

        return modifying_files, deleting_files

    def _dry_run_rsync(self, rsync_command: List[str], delete: bool):
        """Run rsync with --dry-run and warn if files are being deleted.
        """

        # --itemize-changes is needed to detect file changes
        dry_run_options = ["--dry-run", "--itemize-changes"]
        should_warn_delete = False

        if not delete:
            # --delete is needed to detect files that are being deleted in the destination
            should_warn_delete = True
            dry_run_options.append("--delete")

        try:
            result = subprocess.run(
                rsync_command + dry_run_options,
                capture_output=True,
                text=True,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            self._logger.error(f"Error running rsync command: {e}")
            self._logger.error(f">>> stdout: {e.stdout}")
            self._logger.error(f">>> stderr: {e.stderr}")
            raise RuntimeError(f"Rsync failed with return code {e.returncode}")

        _, deleting_files = self._parse_rsync_dry_run_output(result.stdout)

        if should_warn_delete and len(deleting_files):
            click.echo(
                "Detected files that exist in the destination but not in the source. The files will not be deleted by default. You can add '--delete' option to delete the files:"
            )
            click.echo(
                "\n".join([click.style(file, fg="red") for file in deleting_files])
            )

    def _run_rsync(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        local_dir: Optional[str] = None,
        rsync_args: Optional[List[str]] = None,
        is_pull: bool = False,
        include_git_state: bool = False,
        delete: bool = False,
    ):
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )

        default_dir_name = self.get_default_dir_name(id=workspace_model.id)

        local_dir = local_dir or os.getcwd()

        with tempfile.TemporaryDirectory() as tmp_dir:
            host_name, config_file = self.generate_ssh_config_file(
                id=workspace_model.id, ssh_config_path=tmp_dir
            )

            if is_pull:
                source = f"ray@{host_name}:~/{default_dir_name}/"
                destination = local_dir
            else:
                # Ensure source directory has trailing slash for consistent rsync behavior
                # This ensures we sync the contents of the directory, not the directory itself
                source = os.path.join(local_dir, "")  # Ensures trailing slash
                destination = f"ray@{host_name}:~/{default_dir_name}/"

            rsync_args = rsync_args or []

            # exclude .git/objects/info/alternates since we will repack the git repo if needed
            # exclude .anyscale.yaml for legacy reasons
            rsync_args += [
                "--exclude",
                ".git/objects/info/alternates",
                "--exclude",
                ".anyscale.yaml",
            ]

            # repack git repos if needed
            if include_git_state and is_pull:
                self.run_command(
                    id=workspace_model.id,
                    command=f"cd ~/{default_dir_name} && python -m snapshot_util repack_git_repos",
                )
            elif not include_git_state:
                rsync_args += ["--exclude", ".git"]

            # Use -c (--checksum) to avoid retransmitting files that haven't changed
            args = [
                "rsync",
                "-rzlc",
                "-e",
                f"ssh -F {config_file} {subprocess.list2cmdline(ANYSCALE_WORKSPACES_SSH_OPTIONS)}",
                source,
                destination,
            ]

            if delete:
                # --delete removes files in the destination that are not in the source
                # Note: This preserves excluded files (like .git) and is compatible with all rsync implementations
                args.append("--delete")

            if rsync_args:
                args.extend(rsync_args)

            self._dry_run_rsync(args, delete)

            # Add -v / --verbose to the rsync command to be explicit about what is being transferred
            args += ["-v"]

            try:
                subprocess.run(args, check=True, capture_output=True, text=True)
            except subprocess.CalledProcessError as e:
                self._logger.error(f">>> Error running rsync command: {e}")
                self._logger.error(f">>> stdout: {e.stdout}")
                self._logger.error(f">>> stderr: {e.stderr}")
                raise RuntimeError(f"Rsync failed with return code {e.returncode}")

    def pull(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        local_dir: Optional[str] = None,
        pull_git_state: bool = False,
        rsync_args: Optional[List[str]] = None,
        delete: bool = False,
    ):
        self._run_rsync(
            id=id,
            name=name,
            cloud=cloud,
            project=project,
            local_dir=local_dir,
            rsync_args=rsync_args,
            is_pull=True,
            include_git_state=pull_git_state,
            delete=delete,
        )

    def push(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        local_dir: Optional[str] = None,
        push_git_state: bool = False,
        rsync_args: Optional[List[str]] = None,
        delete: bool = False,
    ):
        self._run_rsync(
            id=id,
            name=name,
            cloud=cloud,
            project=project,
            local_dir=local_dir,
            is_pull=False,
            include_git_state=push_git_state,
            rsync_args=rsync_args,
            delete=delete,
        )

    def update(
        self, *, id: Optional[str], config: UpdateWorkspaceConfig  # noqa: A002
    ) -> str:
        workspace = self.client.get_workspace(id=id)  # type: ignore

        if not workspace:
            raise ValueError(f"Workspace with id '{id}' was not found.")

        current_status = self._get_workspace_status(id)
        if current_status != WorkspaceState.TERMINATED:
            raise ValueError(
                "Workspace must be in the TERMINATED state to be updated. Use `anyscale workspace terminate` to terminate the workspace, and `anyscale workspace wait` to wait for the workspace to terminate."
            )

        name = config.name or workspace.name

        compute_config_id = None
        if config.compute_config:
            compute_config_id, _ = self.resolve_compute_config_and_cloud_id(
                compute_config=config.compute_config, cloud=None,  # type: ignore
            )

        build_id = None
        if config.containerfile is not None:
            build_id = self._image_sdk.build_image_from_containerfile(
                name=f"image-for-workspace-{name}",
                containerfile=self.get_containerfile_contents(config.containerfile),
                ray_version=config.ray_version,
            )
        elif config.image_uri is not None:
            build_id = self._image_sdk.registery_image(
                image_uri=config.image_uri,
                registry_login_secret=config.registry_login_secret,
                ray_version=config.ray_version,
            )

        dynamic_requirements = None
        if (
            config.requirements
            and self._image_sdk.enable_image_build_for_tracked_requirements
        ):
            requirements = (
                parse_requirements_file(config.requirements)
                if isinstance(config.requirements, str)
                else config.requirements
            )
            if requirements:
                build_id = self._image_sdk.build_image_from_requirements(
                    name=f"image-for-workspace-{name}",
                    base_build_id=self.client.get_default_build_id(),
                    requirements=requirements,
                )
        elif config.requirements:
            dynamic_requirements = (
                parse_requirements_file(config.requirements)
                if isinstance(config.requirements, str)
                else config.requirements
            )

        self.client.update_workspace(
            workspace_id=id,
            name=config.name,
            compute_config_id=compute_config_id,
            cluster_environment_build_id=build_id,
            idle_timeout_minutes=config.idle_termination_minutes,
        )

        self._logger.info(f"Workspace updated successfully id: {id}")

        if dynamic_requirements:
            self.client.update_workspace_dependencies_offline_only(
                workspace_id=id, requirements=dynamic_requirements
            )
            self._logger.info(f"Applied dynamic requirements to workspace id: {id}")
        if config.env_vars:
            self.client.update_workspace_env_vars_offline_only(
                workspace_id=id, env_vars=config.env_vars
            )
            self._logger.info(f"Applied environment variables to workspace id: {id}")

        return id  # type: ignore

    def _convert_cluster_state_to_workspace_state(
        self, state: SessionState
    ) -> WorkspaceState:
        return cast(
            WorkspaceState,
            self._BACKEND_SESSION_STATE_TO_WORKSPACE_STATE.get(  # type: ignore
                state, WorkspaceState.UNKNOWN
            ),
        )

    def _convert_env_var_list_to_dict(
        self, env_vars: Optional[List[str]]
    ) -> Dict[str, str]:
        if not env_vars:
            return {}
        return dict([env_var.split("=", 1) for env_var in env_vars])

    def _convert_requirements_str_to_list(
        self, requirements: Optional[str]
    ) -> List[str]:
        if not requirements:
            return []
        return [req for req in requirements.split("\n") if req]

    def _to_workspace(
        self, workspace: ExperimentalWorkspace, *, include_config: bool = True
    ) -> Workspace:
        """Convert ExperimentalWorkspace to Workspace.

        Args:
            workspace: Internal workspace model from API
            include_config: If True, fetch full config (expensive). If False, only metadata.
        """
        state = self._BACKEND_SESSION_STATE_TO_WORKSPACE_STATE.get(
            workspace.state, WorkspaceState.UNKNOWN
        )

        # Optionally fetch expensive config data
        config = self._fetch_workspace_config(workspace) if include_config else None

        return Workspace(
            id=workspace.id or "",
            name=workspace.name or "",
            state=state,
            project_id=workspace.project_id,
            cloud_id=workspace.cloud_id,
            creator_id=workspace.creator_id,
            creator_email=workspace.creator_email,
            created_at=workspace.created_at,
            last_started_at=workspace.latest_started_at,
            cluster_id=workspace.cluster_id,
            config=config,
        )

    def _fetch_workspace_config(
        self, workspace: ExperimentalWorkspace
    ) -> WorkspaceConfig:
        """Fetch full workspace configuration (expensive, makes multiple API calls).

        Optimizes performance by parallelizing independent API calls using ThreadPoolExecutor.
        ThreadPoolExecutor is appropriate here because the OpenAPI clients use synchronous I/O
        (requests library), not async/await.
        """
        # First fetch cluster (needed for subsequent calls)
        cluster = self.client.get_workspace_cluster(workspace.id)
        if not cluster:
            raise ValueError(
                f"Workspace cluster with ID '{workspace.cluster_id}' was not found."
            )

        # Parallelize all independent API calls
        with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
            # Submit all fetches in parallel - more Pythonic dict pattern
            futures = {
                "image_uri": executor.submit(
                    self._image_sdk.get_image_uri_from_build_id, cluster.build_id
                ),
                "image_build": executor.submit(
                    self._image_sdk.get_image_build, cluster.build_id
                ),
                "compute": executor.submit(
                    self.get_user_facing_compute_config, workspace.compute_config_id
                ),
                "cloud": executor.submit(
                    self.client.get_cloud, cloud_id=workspace.cloud_id
                ),
                "project": executor.submit(
                    self.client.get_project, project_id=workspace.project_id
                ),
                "artifacts": executor.submit(
                    self.client.get_workspace_proxied_dataplane_artifacts,
                    workspace_id=workspace.id,
                ),
            }

            # Collect all results - will raise exception if any call failed
            results = {key: future.result() for key, future in futures.items()}

        # Validate required results
        if not results["image_uri"]:
            raise ValueError(f"Failed to get image URI for build {cluster.build_id}.")
        if not results["image_build"]:
            raise ValueError(f"Failed to get image build {cluster.build_id}.")

        env_vars = self._convert_env_var_list_to_dict(
            results["artifacts"].environment_variables
        )
        requirements = self._convert_requirements_str_to_list(
            results["artifacts"].requirements
        )

        return WorkspaceConfig(
            name=workspace.name,
            compute_config=results["compute"],
            registry_login_secret=results["image_build"].registry_login_secret,
            image_uri=str(results["image_uri"]),
            requirements=requirements if requirements else None,
            idle_termination_minutes=cluster.idle_timeout,
            env_vars=env_vars,
            project=results["project"].name if results["project"] else None,
            cloud=results["cloud"].name if results["cloud"] else None,
        )

    def get(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        include_config: bool = True,
    ) -> Workspace:
        """Get a workspace by name or ID.

        Args:
            include_config: If True, fetch full config (default). Set to False for efficiency.
        """
        workspace = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        return self._to_workspace(workspace, include_config=include_config)

    def list(  # noqa: PLR0913, PLR0917
        self,
        *,
        workspace_id: Optional[str] = None,
        name: Optional[str] = None,
        project: Optional[str] = None,
        cloud: Optional[str] = None,
        creator_id: Optional[str] = None,
        state_filter: Optional[Union[List[WorkspaceState], List[str]]] = None,
        tags_filter: Optional[Dict[str, List[str]]] = None,
        include_config: bool = False,
        sort_field: Optional[Union[str, WorkspaceSortField]] = None,
        sort_order: Optional[Union[str, WorkspaceSortOrder]] = None,
        max_items: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ResultIterator[Workspace]:
        """List workspaces with optional filters.

        Args:
            state_filter: List of states to include. May be WorkspaceState enums or case-insensitive strings.
                         If None, defaults to non-terminated states.
            tags_filter: Filter by tags. Dict mapping tag keys to lists of values.
                        Tags with the same key are ORed, different keys are ANDed.
            include_config: If True, fetch full config for each workspace (expensive, makes multiple API calls).
                           Defaults to False for efficiency. Set to True when full config is needed.
            sort_field: Field to sort by. Defaults to status (active first) then created_at (desc).
            sort_order: Sort order (ASC or DESC). Defaults to appropriate order for sort_field.
        """
        MAX_PAGE_SIZE = 100

        if page_size is not None and not (1 <= page_size <= MAX_PAGE_SIZE):
            raise ValueError(
                f"page_size must be between 1 and {MAX_PAGE_SIZE}, inclusive."
            )

        # Resolve cloud and project to IDs
        cloud_id = self.client.get_cloud_id(cloud_name=cloud) if cloud else None
        project_id = (
            self.client.get_project_id(parent_cloud_id=cloud_id, name=project)
            if project
            else None
        )

        # Normalize state filter
        # Backend expects SessionState enum values as strings
        normalized_states = _normalize_state_filter_to_backend(state_filter)

        # Convert tags_filter dict to backend format (list of "key:value" strings)
        tag_filter_list: Optional[List[str]] = None
        if tags_filter:
            tag_filter_list = []
            for key, values in tags_filter.items():
                for value in values:
                    tag_filter_list.append(f"{key}:{value}")

        # Handle single workspace lookup by ID
        if workspace_id is not None:
            workspace = self.client.get_workspace(id=workspace_id)

            if workspace is None:
                # Return empty iterator
                return ResultIterator(
                    page_token=None,
                    max_items=0,
                    fetch_page=lambda _: SimpleNamespace(
                        results=[], metadata=SimpleNamespace(next_paging_token=None),
                    ),
                    parse_fn=None,
                )

            # Return single-item iterator
            summary = self._to_workspace(workspace, include_config=include_config)
            return ResultIterator(
                page_token=None,
                max_items=1,
                fetch_page=lambda token: SimpleNamespace(
                    results=[summary] if token is None else [],
                    metadata=SimpleNamespace(next_paging_token=None),
                ),
                parse_fn=None,
            )

        # Handle paginated list
        def _fetch_page(token: Optional[str]) -> SimpleNamespace:
            # Use internal API client directly as there's no interface method for list_workspaces yet
            response = self.client._internal_api_client.list_workspaces_api_v2_experimental_workspaces_get(  # noqa: SLF001
                project_id=project_id,
                cloud_id=cloud_id,
                name=name,
                creator_id=creator_id,
                state_filter=normalized_states,
                tag_filter=tag_filter_list,
                sort_field=sort_field,
                sort_order=sort_order,
                count=page_size or 50,
                paging_token=token,
            )

            results = response.results if response.results else []
            next_token = (
                response.metadata.next_paging_token if response.metadata else None
            )

            return SimpleNamespace(
                results=results, metadata=SimpleNamespace(next_paging_token=next_token),
            )

        return ResultIterator(
            page_token=None,
            max_items=max_items,
            fetch_page=_fetch_page,
            parse_fn=lambda ws: self._to_workspace(ws, include_config=include_config),
        )

    def add_tags(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        tags: Dict[str, str],
    ) -> None:
        """Upsert (add/update) tag key/value pairs for a workspace."""
        if not tags:
            raise ValueError("At least one tag must be provided.")

        if id is not None:
            resource_id = id
        else:
            if name is None and not self.client.inside_workspace():
                raise ValueError(
                    "Either 'id' or 'name' must be provided when running outside of a workspace."
                )
            model = self._resolve_to_workspace_model(
                id=id, name=name, cloud=cloud, project=project
            )
            resource_id = model.id  # type: ignore

        self.client.upsert_resource_tags(
            ResourceTagResourceType.WORKSPACE, resource_id, tags
        )

    def remove_tags(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        keys: List[str],
    ) -> None:
        """Remove tag keys from a workspace."""
        if not keys:
            raise ValueError("At least one tag key must be provided.")

        if id is not None:
            resource_id = id
        else:
            if name is None and not self.client.inside_workspace():
                raise ValueError(
                    "Either 'id' or 'name' must be provided when running outside of a workspace."
                )
            model = self._resolve_to_workspace_model(
                id=id, name=name, cloud=cloud, project=project
            )
            resource_id = model.id  # type: ignore

        self.client.delete_resource_tags(
            ResourceTagResourceType.WORKSPACE, resource_id, keys
        )

    def list_tags(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> Dict[str, str]:
        """List tags for a workspace as a key/value mapping."""
        if id is not None:
            resource_id = id
        else:
            model = self._resolve_to_workspace_model(
                id=id, name=name, cloud=cloud, project=project
            )
            resource_id = model.id  # type: ignore

        records = self.client.list_resource_tags(
            ResourceTagResourceType.WORKSPACE, resource_id
        )
        return {r.key: r.value for r in records if r and r.key is not None}


def _normalize_state_filter_to_backend(
    states: Optional[Union[List[WorkspaceState], List[str]]],
) -> Optional[List[str]]:
    """Normalize frontend WorkspaceState filter to backend SessionState values.

    Maps frontend WorkspaceState enums to their corresponding backend SessionState values.
    None or empty list: no filtering (includes all states).
    """
    # Map WorkspaceState to backend SessionState values
    WORKSPACE_TO_SESSION_STATES: Dict[WorkspaceState, List[str]] = {
        WorkspaceState.STARTING: [
            "StartingUp",
            "AwaitingStartup",
            "AwaitingFileMounts",
        ],
        WorkspaceState.RUNNING: ["Running"],
        WorkspaceState.UPDATING: ["Updating"],
        WorkspaceState.TERMINATING: ["Stopping", "Terminating"],
        WorkspaceState.TERMINATED: ["Stopped", "Terminated"],
        WorkspaceState.ERRORED: [
            "UpdatingErrored",
            "StartupErrored",
            "TerminatingErrored",
            "StoppingErrored",
        ],
        WorkspaceState.UNKNOWN: [],
    }

    # None or empty list means no filtering
    if states is None or len(states) == 0:
        return None

    backend_states: List[str] = []
    for s in states:
        if isinstance(s, WorkspaceState):
            backend_states.extend(WORKSPACE_TO_SESSION_STATES.get(s, []))
        elif isinstance(s, str):
            backend_states.append(s)
        else:
            raise TypeError(
                "'state_filter' entries must be WorkspaceState or str, "
                f"got {type(s).__name__}"
            )
    return backend_states if backend_states else None



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class WorkspaceTemplate(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'description': 'str',
        'mins_to_complete': 'int',
        'icon_type': 'WorkspaceTemplateIconType',
        'icon_bg_color': 'WorkspaceTemplateIconBackgroundColor',
        'complexity': 'WorkspaceTemplateComplexity',
        'category': 'WorkspaceTemplateCategory',
        'labels': 'list[str]',
        'published_version_id': 'str',
        'cloud_id': 'str',
        'is_global': 'bool',
        'id': 'str',
        'creator_id': 'str',
        'created_at': 'datetime',
        'updated_at': 'datetime',
        'organization_id': 'str',
        'creator_email': 'str',
        'current_version': 'WorkspaceTemplateVersionDataObject',
        'version': 'int'
    }

    attribute_map = {
        'name': 'name',
        'description': 'description',
        'mins_to_complete': 'mins_to_complete',
        'icon_type': 'icon_type',
        'icon_bg_color': 'icon_bg_color',
        'complexity': 'complexity',
        'category': 'category',
        'labels': 'labels',
        'published_version_id': 'published_version_id',
        'cloud_id': 'cloud_id',
        'is_global': 'is_global',
        'id': 'id',
        'creator_id': 'creator_id',
        'created_at': 'created_at',
        'updated_at': 'updated_at',
        'organization_id': 'organization_id',
        'creator_email': 'creator_email',
        'current_version': 'current_version',
        'version': 'version'
    }

    def __init__(self, name=None, description=None, mins_to_complete=None, icon_type=None, icon_bg_color=None, complexity=None, category=None, labels=None, published_version_id=None, cloud_id=None, is_global=False, id=None, creator_id=None, created_at=None, updated_at=None, organization_id=None, creator_email=None, current_version=None, version=None, local_vars_configuration=None):  # noqa: E501
        """WorkspaceTemplate - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._description = None
        self._mins_to_complete = None
        self._icon_type = None
        self._icon_bg_color = None
        self._complexity = None
        self._category = None
        self._labels = None
        self._published_version_id = None
        self._cloud_id = None
        self._is_global = None
        self._id = None
        self._creator_id = None
        self._created_at = None
        self._updated_at = None
        self._organization_id = None
        self._creator_email = None
        self._current_version = None
        self._version = None
        self.discriminator = None

        self.name = name
        if description is not None:
            self.description = description
        if mins_to_complete is not None:
            self.mins_to_complete = mins_to_complete
        if icon_type is not None:
            self.icon_type = icon_type
        if icon_bg_color is not None:
            self.icon_bg_color = icon_bg_color
        if complexity is not None:
            self.complexity = complexity
        if category is not None:
            self.category = category
        if labels is not None:
            self.labels = labels
        if published_version_id is not None:
            self.published_version_id = published_version_id
        self.cloud_id = cloud_id
        if is_global is not None:
            self.is_global = is_global
        self.id = id
        self.creator_id = creator_id
        self.created_at = created_at
        if updated_at is not None:
            self.updated_at = updated_at
        self.organization_id = organization_id
        if creator_email is not None:
            self.creator_email = creator_email
        if current_version is not None:
            self.current_version = current_version
        if version is not None:
            self.version = version

    @property
    def name(self):
        """Gets the name of this WorkspaceTemplate.  # noqa: E501

        Name of the template  # noqa: E501

        :return: The name of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this WorkspaceTemplate.

        Name of the template  # noqa: E501

        :param name: The name of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def description(self):
        """Gets the description of this WorkspaceTemplate.  # noqa: E501

        Description of the template  # noqa: E501

        :return: The description of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this WorkspaceTemplate.

        Description of the template  # noqa: E501

        :param description: The description of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def mins_to_complete(self):
        """Gets the mins_to_complete of this WorkspaceTemplate.  # noqa: E501

        Estimated minutes to complete this template  # noqa: E501

        :return: The mins_to_complete of this WorkspaceTemplate.  # noqa: E501
        :rtype: int
        """
        return self._mins_to_complete

    @mins_to_complete.setter
    def mins_to_complete(self, mins_to_complete):
        """Sets the mins_to_complete of this WorkspaceTemplate.

        Estimated minutes to complete this template  # noqa: E501

        :param mins_to_complete: The mins_to_complete of this WorkspaceTemplate.  # noqa: E501
        :type: int
        """

        self._mins_to_complete = mins_to_complete

    @property
    def icon_type(self):
        """Gets the icon_type of this WorkspaceTemplate.  # noqa: E501

        Icon type for the template  # noqa: E501

        :return: The icon_type of this WorkspaceTemplate.  # noqa: E501
        :rtype: WorkspaceTemplateIconType
        """
        return self._icon_type

    @icon_type.setter
    def icon_type(self, icon_type):
        """Sets the icon_type of this WorkspaceTemplate.

        Icon type for the template  # noqa: E501

        :param icon_type: The icon_type of this WorkspaceTemplate.  # noqa: E501
        :type: WorkspaceTemplateIconType
        """

        self._icon_type = icon_type

    @property
    def icon_bg_color(self):
        """Gets the icon_bg_color of this WorkspaceTemplate.  # noqa: E501

        Background color for the icon  # noqa: E501

        :return: The icon_bg_color of this WorkspaceTemplate.  # noqa: E501
        :rtype: WorkspaceTemplateIconBackgroundColor
        """
        return self._icon_bg_color

    @icon_bg_color.setter
    def icon_bg_color(self, icon_bg_color):
        """Sets the icon_bg_color of this WorkspaceTemplate.

        Background color for the icon  # noqa: E501

        :param icon_bg_color: The icon_bg_color of this WorkspaceTemplate.  # noqa: E501
        :type: WorkspaceTemplateIconBackgroundColor
        """

        self._icon_bg_color = icon_bg_color

    @property
    def complexity(self):
        """Gets the complexity of this WorkspaceTemplate.  # noqa: E501

        Complexity level of the template  # noqa: E501

        :return: The complexity of this WorkspaceTemplate.  # noqa: E501
        :rtype: WorkspaceTemplateComplexity
        """
        return self._complexity

    @complexity.setter
    def complexity(self, complexity):
        """Sets the complexity of this WorkspaceTemplate.

        Complexity level of the template  # noqa: E501

        :param complexity: The complexity of this WorkspaceTemplate.  # noqa: E501
        :type: WorkspaceTemplateComplexity
        """

        self._complexity = complexity

    @property
    def category(self):
        """Gets the category of this WorkspaceTemplate.  # noqa: E501

        Category of the template  # noqa: E501

        :return: The category of this WorkspaceTemplate.  # noqa: E501
        :rtype: WorkspaceTemplateCategory
        """
        return self._category

    @category.setter
    def category(self, category):
        """Sets the category of this WorkspaceTemplate.

        Category of the template  # noqa: E501

        :param category: The category of this WorkspaceTemplate.  # noqa: E501
        :type: WorkspaceTemplateCategory
        """

        self._category = category

    @property
    def labels(self):
        """Gets the labels of this WorkspaceTemplate.  # noqa: E501

        Searchable labels for the template  # noqa: E501

        :return: The labels of this WorkspaceTemplate.  # noqa: E501
        :rtype: list[str]
        """
        return self._labels

    @labels.setter
    def labels(self, labels):
        """Sets the labels of this WorkspaceTemplate.

        Searchable labels for the template  # noqa: E501

        :param labels: The labels of this WorkspaceTemplate.  # noqa: E501
        :type: list[str]
        """

        self._labels = labels

    @property
    def published_version_id(self):
        """Gets the published_version_id of this WorkspaceTemplate.  # noqa: E501

        ID of the published version  # noqa: E501

        :return: The published_version_id of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._published_version_id

    @published_version_id.setter
    def published_version_id(self, published_version_id):
        """Sets the published_version_id of this WorkspaceTemplate.

        ID of the published version  # noqa: E501

        :param published_version_id: The published_version_id of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """

        self._published_version_id = published_version_id

    @property
    def cloud_id(self):
        """Gets the cloud_id of this WorkspaceTemplate.  # noqa: E501

        Cloud ID this template is associated with  # noqa: E501

        :return: The cloud_id of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this WorkspaceTemplate.

        Cloud ID this template is associated with  # noqa: E501

        :param cloud_id: The cloud_id of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def is_global(self):
        """Gets the is_global of this WorkspaceTemplate.  # noqa: E501

        Whether this is a global template  # noqa: E501

        :return: The is_global of this WorkspaceTemplate.  # noqa: E501
        :rtype: bool
        """
        return self._is_global

    @is_global.setter
    def is_global(self, is_global):
        """Sets the is_global of this WorkspaceTemplate.

        Whether this is a global template  # noqa: E501

        :param is_global: The is_global of this WorkspaceTemplate.  # noqa: E501
        :type: bool
        """

        self._is_global = is_global

    @property
    def id(self):
        """Gets the id of this WorkspaceTemplate.  # noqa: E501

        ID of the template  # noqa: E501

        :return: The id of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this WorkspaceTemplate.

        ID of the template  # noqa: E501

        :param id: The id of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def creator_id(self):
        """Gets the creator_id of this WorkspaceTemplate.  # noqa: E501

        ID of the user who created this template  # noqa: E501

        :return: The creator_id of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this WorkspaceTemplate.

        ID of the user who created this template  # noqa: E501

        :param creator_id: The creator_id of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def created_at(self):
        """Gets the created_at of this WorkspaceTemplate.  # noqa: E501

        When this template was created  # noqa: E501

        :return: The created_at of this WorkspaceTemplate.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this WorkspaceTemplate.

        When this template was created  # noqa: E501

        :param created_at: The created_at of this WorkspaceTemplate.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this WorkspaceTemplate.  # noqa: E501

        When this template was last updated  # noqa: E501

        :return: The updated_at of this WorkspaceTemplate.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this WorkspaceTemplate.

        When this template was last updated  # noqa: E501

        :param updated_at: The updated_at of this WorkspaceTemplate.  # noqa: E501
        :type: datetime
        """

        self._updated_at = updated_at

    @property
    def organization_id(self):
        """Gets the organization_id of this WorkspaceTemplate.  # noqa: E501

        ID of the organization this template is associated with  # noqa: E501

        :return: The organization_id of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this WorkspaceTemplate.

        ID of the organization this template is associated with  # noqa: E501

        :param organization_id: The organization_id of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and organization_id is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_id`, must not be `None`")  # noqa: E501

        self._organization_id = organization_id

    @property
    def creator_email(self):
        """Gets the creator_email of this WorkspaceTemplate.  # noqa: E501

        Email of the creator  # noqa: E501

        :return: The creator_email of this WorkspaceTemplate.  # noqa: E501
        :rtype: str
        """
        return self._creator_email

    @creator_email.setter
    def creator_email(self, creator_email):
        """Sets the creator_email of this WorkspaceTemplate.

        Email of the creator  # noqa: E501

        :param creator_email: The creator_email of this WorkspaceTemplate.  # noqa: E501
        :type: str
        """

        self._creator_email = creator_email

    @property
    def current_version(self):
        """Gets the current_version of this WorkspaceTemplate.  # noqa: E501

        Current version of the template. This is the version being published or the latest version if no version is published.  # noqa: E501

        :return: The current_version of this WorkspaceTemplate.  # noqa: E501
        :rtype: WorkspaceTemplateVersionDataObject
        """
        return self._current_version

    @current_version.setter
    def current_version(self, current_version):
        """Sets the current_version of this WorkspaceTemplate.

        Current version of the template. This is the version being published or the latest version if no version is published.  # noqa: E501

        :param current_version: The current_version of this WorkspaceTemplate.  # noqa: E501
        :type: WorkspaceTemplateVersionDataObject
        """

        self._current_version = current_version

    @property
    def version(self):
        """Gets the version of this WorkspaceTemplate.  # noqa: E501

        Version of the template  # noqa: E501

        :return: The version of this WorkspaceTemplate.  # noqa: E501
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this WorkspaceTemplate.

        Version of the template  # noqa: E501

        :param version: The version of this WorkspaceTemplate.  # noqa: E501
        :type: int
        """

        self._version = version

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, WorkspaceTemplate):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, WorkspaceTemplate):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class CreateProductionJobConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'entrypoint': 'str',
        'ray_serve_config': 'object',
        'runtime_env': 'RayRuntimeEnvConfig',
        'build_id': 'str',
        'image_uri': 'str',
        'registry_login_secret': 'str',
        'compute_config_id': 'str',
        'compute_config': 'CreateClusterComputeConfig',
        'max_retries': 'int',
        'timeout_s': 'int',
        'runtime_env_config': 'RayRuntimeEnvConfig'
    }

    attribute_map = {
        'entrypoint': 'entrypoint',
        'ray_serve_config': 'ray_serve_config',
        'runtime_env': 'runtime_env',
        'build_id': 'build_id',
        'image_uri': 'image_uri',
        'registry_login_secret': 'registry_login_secret',
        'compute_config_id': 'compute_config_id',
        'compute_config': 'compute_config',
        'max_retries': 'max_retries',
        'timeout_s': 'timeout_s',
        'runtime_env_config': 'runtime_env_config'
    }

    def __init__(self, entrypoint='', ray_serve_config=None, runtime_env=None, build_id=None, image_uri=None, registry_login_secret=None, compute_config_id=None, compute_config=None, max_retries=5, timeout_s=None, runtime_env_config=None, local_vars_configuration=None):  # noqa: E501
        """CreateProductionJobConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._entrypoint = None
        self._ray_serve_config = None
        self._runtime_env = None
        self._build_id = None
        self._image_uri = None
        self._registry_login_secret = None
        self._compute_config_id = None
        self._compute_config = None
        self._max_retries = None
        self._timeout_s = None
        self._runtime_env_config = None
        self.discriminator = None

        if entrypoint is not None:
            self.entrypoint = entrypoint
        if ray_serve_config is not None:
            self.ray_serve_config = ray_serve_config
        if runtime_env is not None:
            self.runtime_env = runtime_env
        if build_id is not None:
            self.build_id = build_id
        if image_uri is not None:
            self.image_uri = image_uri
        if registry_login_secret is not None:
            self.registry_login_secret = registry_login_secret
        if compute_config_id is not None:
            self.compute_config_id = compute_config_id
        if compute_config is not None:
            self.compute_config = compute_config
        if max_retries is not None:
            self.max_retries = max_retries
        if timeout_s is not None:
            self.timeout_s = timeout_s
        if runtime_env_config is not None:
            self.runtime_env_config = runtime_env_config

    @property
    def entrypoint(self):
        """Gets the entrypoint of this CreateProductionJobConfig.  # noqa: E501

        A script that will be run to start your job.This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'  # noqa: E501

        :return: The entrypoint of this CreateProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._entrypoint

    @entrypoint.setter
    def entrypoint(self, entrypoint):
        """Sets the entrypoint of this CreateProductionJobConfig.

        A script that will be run to start your job.This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'  # noqa: E501

        :param entrypoint: The entrypoint of this CreateProductionJobConfig.  # noqa: E501
        :type: str
        """

        self._entrypoint = entrypoint

    @property
    def ray_serve_config(self):
        """Gets the ray_serve_config of this CreateProductionJobConfig.  # noqa: E501

        The Ray Serve config to use for this Production service. This config defines your Ray Serve application, and will be passed directly to Ray Serve. You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html  # noqa: E501

        :return: The ray_serve_config of this CreateProductionJobConfig.  # noqa: E501
        :rtype: object
        """
        return self._ray_serve_config

    @ray_serve_config.setter
    def ray_serve_config(self, ray_serve_config):
        """Sets the ray_serve_config of this CreateProductionJobConfig.

        The Ray Serve config to use for this Production service. This config defines your Ray Serve application, and will be passed directly to Ray Serve. You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html  # noqa: E501

        :param ray_serve_config: The ray_serve_config of this CreateProductionJobConfig.  # noqa: E501
        :type: object
        """

        self._ray_serve_config = ray_serve_config

    @property
    def runtime_env(self):
        """Gets the runtime_env of this CreateProductionJobConfig.  # noqa: E501

        A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.  # noqa: E501

        :return: The runtime_env of this CreateProductionJobConfig.  # noqa: E501
        :rtype: RayRuntimeEnvConfig
        """
        return self._runtime_env

    @runtime_env.setter
    def runtime_env(self, runtime_env):
        """Sets the runtime_env of this CreateProductionJobConfig.

        A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.  # noqa: E501

        :param runtime_env: The runtime_env of this CreateProductionJobConfig.  # noqa: E501
        :type: RayRuntimeEnvConfig
        """

        self._runtime_env = runtime_env

    @property
    def build_id(self):
        """Gets the build_id of this CreateProductionJobConfig.  # noqa: E501

        The id of the cluster env build. This id will determine the docker image your job is run on. Either build_id or image_uri must be provided.  # noqa: E501

        :return: The build_id of this CreateProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._build_id

    @build_id.setter
    def build_id(self, build_id):
        """Sets the build_id of this CreateProductionJobConfig.

        The id of the cluster env build. This id will determine the docker image your job is run on. Either build_id or image_uri must be provided.  # noqa: E501

        :param build_id: The build_id of this CreateProductionJobConfig.  # noqa: E501
        :type: str
        """

        self._build_id = build_id

    @property
    def image_uri(self):
        """Gets the image_uri of this CreateProductionJobConfig.  # noqa: E501

        The image URI of the container in which the job will run. Can be an Anyscale image (anyscale/ray:*), cluster environment image (anyscale/image/*), or a BYOD image from an external registry. Either build_id or image_uri must be provided.  # noqa: E501

        :return: The image_uri of this CreateProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._image_uri

    @image_uri.setter
    def image_uri(self, image_uri):
        """Sets the image_uri of this CreateProductionJobConfig.

        The image URI of the container in which the job will run. Can be an Anyscale image (anyscale/ray:*), cluster environment image (anyscale/image/*), or a BYOD image from an external registry. Either build_id or image_uri must be provided.  # noqa: E501

        :param image_uri: The image_uri of this CreateProductionJobConfig.  # noqa: E501
        :type: str
        """

        self._image_uri = image_uri

    @property
    def registry_login_secret(self):
        """Gets the registry_login_secret of this CreateProductionJobConfig.  # noqa: E501

        The name or identifier of a secret containing credentials to authenticate to the docker registry hosting the image. Only used with image_uri for BYOD images from external registries.  # noqa: E501

        :return: The registry_login_secret of this CreateProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._registry_login_secret

    @registry_login_secret.setter
    def registry_login_secret(self, registry_login_secret):
        """Sets the registry_login_secret of this CreateProductionJobConfig.

        The name or identifier of a secret containing credentials to authenticate to the docker registry hosting the image. Only used with image_uri for BYOD images from external registries.  # noqa: E501

        :param registry_login_secret: The registry_login_secret of this CreateProductionJobConfig.  # noqa: E501
        :type: str
        """

        self._registry_login_secret = registry_login_secret

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this CreateProductionJobConfig.  # noqa: E501

        The id of the compute configuration that you want to use. This id will specify the resources required for your job  # noqa: E501

        :return: The compute_config_id of this CreateProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this CreateProductionJobConfig.

        The id of the compute configuration that you want to use. This id will specify the resources required for your job  # noqa: E501

        :param compute_config_id: The compute_config_id of this CreateProductionJobConfig.  # noqa: E501
        :type: str
        """

        self._compute_config_id = compute_config_id

    @property
    def compute_config(self):
        """Gets the compute_config of this CreateProductionJobConfig.  # noqa: E501

        One-off compute that the cluster will use.  # noqa: E501

        :return: The compute_config of this CreateProductionJobConfig.  # noqa: E501
        :rtype: CreateClusterComputeConfig
        """
        return self._compute_config

    @compute_config.setter
    def compute_config(self, compute_config):
        """Sets the compute_config of this CreateProductionJobConfig.

        One-off compute that the cluster will use.  # noqa: E501

        :param compute_config: The compute_config of this CreateProductionJobConfig.  # noqa: E501
        :type: CreateClusterComputeConfig
        """

        self._compute_config = compute_config

    @property
    def max_retries(self):
        """Gets the max_retries of this CreateProductionJobConfig.  # noqa: E501

        The number of retries this job will attempt on failure. Set to None to set infinite retries  # noqa: E501

        :return: The max_retries of this CreateProductionJobConfig.  # noqa: E501
        :rtype: int
        """
        return self._max_retries

    @max_retries.setter
    def max_retries(self, max_retries):
        """Sets the max_retries of this CreateProductionJobConfig.

        The number of retries this job will attempt on failure. Set to None to set infinite retries  # noqa: E501

        :param max_retries: The max_retries of this CreateProductionJobConfig.  # noqa: E501
        :type: int
        """

        self._max_retries = max_retries

    @property
    def timeout_s(self):
        """Gets the timeout_s of this CreateProductionJobConfig.  # noqa: E501

        The timeout in seconds for each job run. Set to None for no limit to be set  # noqa: E501

        :return: The timeout_s of this CreateProductionJobConfig.  # noqa: E501
        :rtype: int
        """
        return self._timeout_s

    @timeout_s.setter
    def timeout_s(self, timeout_s):
        """Sets the timeout_s of this CreateProductionJobConfig.

        The timeout in seconds for each job run. Set to None for no limit to be set  # noqa: E501

        :param timeout_s: The timeout_s of this CreateProductionJobConfig.  # noqa: E501
        :type: int
        """

        self._timeout_s = timeout_s

    @property
    def runtime_env_config(self):
        """Gets the runtime_env_config of this CreateProductionJobConfig.  # noqa: E501

        DEPRECATED: Use runtime_env  # noqa: E501

        :return: The runtime_env_config of this CreateProductionJobConfig.  # noqa: E501
        :rtype: RayRuntimeEnvConfig
        """
        return self._runtime_env_config

    @runtime_env_config.setter
    def runtime_env_config(self, runtime_env_config):
        """Sets the runtime_env_config of this CreateProductionJobConfig.

        DEPRECATED: Use runtime_env  # noqa: E501

        :param runtime_env_config: The runtime_env_config of this CreateProductionJobConfig.  # noqa: E501
        :type: RayRuntimeEnvConfig
        """

        self._runtime_env_config = runtime_env_config

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateProductionJobConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateProductionJobConfig):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedProductionJob(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'description': 'str',
        'created_at': 'datetime',
        'creator_id': 'str',
        'config': 'ProductionJobConfig',
        'job_queue_config': 'JobQueueConfig',
        'state': 'DecoratedProductionJobStateTransition',
        'project_id': 'str',
        'last_job_run_id': 'str',
        'schedule_id': 'str',
        'job_queue_id': 'str',
        'overview_url': 'str',
        'is_service': 'bool',
        'url': 'str',
        'token': 'str',
        'access': 'UserServiceAccessTypes',
        'healthcheck_url': 'str',
        'archived_at': 'datetime',
        'cloud_id': 'str',
        'project': 'MiniProject',
        'creator': 'MiniUser',
        'last_job_run': 'MiniJobRun',
        'schedule': 'MiniSchedule',
        'integration_details': 'IntegrationDetails'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'description': 'description',
        'created_at': 'created_at',
        'creator_id': 'creator_id',
        'config': 'config',
        'job_queue_config': 'job_queue_config',
        'state': 'state',
        'project_id': 'project_id',
        'last_job_run_id': 'last_job_run_id',
        'schedule_id': 'schedule_id',
        'job_queue_id': 'job_queue_id',
        'overview_url': 'overview_url',
        'is_service': 'is_service',
        'url': 'url',
        'token': 'token',
        'access': 'access',
        'healthcheck_url': 'healthcheck_url',
        'archived_at': 'archived_at',
        'cloud_id': 'cloud_id',
        'project': 'project',
        'creator': 'creator',
        'last_job_run': 'last_job_run',
        'schedule': 'schedule',
        'integration_details': 'integration_details'
    }

    def __init__(self, id=None, name=None, description=None, created_at=None, creator_id=None, config=None, job_queue_config=None, state=None, project_id=None, last_job_run_id=None, schedule_id=None, job_queue_id=None, overview_url=None, is_service=None, url=None, token=None, access=None, healthcheck_url=None, archived_at=None, cloud_id=None, project=None, creator=None, last_job_run=None, schedule=None, integration_details=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedProductionJob - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._description = None
        self._created_at = None
        self._creator_id = None
        self._config = None
        self._job_queue_config = None
        self._state = None
        self._project_id = None
        self._last_job_run_id = None
        self._schedule_id = None
        self._job_queue_id = None
        self._overview_url = None
        self._is_service = None
        self._url = None
        self._token = None
        self._access = None
        self._healthcheck_url = None
        self._archived_at = None
        self._cloud_id = None
        self._project = None
        self._creator = None
        self._last_job_run = None
        self._schedule = None
        self._integration_details = None
        self.discriminator = None

        self.id = id
        self.name = name
        if description is not None:
            self.description = description
        self.created_at = created_at
        self.creator_id = creator_id
        self.config = config
        if job_queue_config is not None:
            self.job_queue_config = job_queue_config
        self.state = state
        self.project_id = project_id
        if last_job_run_id is not None:
            self.last_job_run_id = last_job_run_id
        if schedule_id is not None:
            self.schedule_id = schedule_id
        if job_queue_id is not None:
            self.job_queue_id = job_queue_id
        self.overview_url = overview_url
        self.is_service = is_service
        if url is not None:
            self.url = url
        if token is not None:
            self.token = token
        if access is not None:
            self.access = access
        if healthcheck_url is not None:
            self.healthcheck_url = healthcheck_url
        if archived_at is not None:
            self.archived_at = archived_at
        self.cloud_id = cloud_id
        self.project = project
        self.creator = creator
        if last_job_run is not None:
            self.last_job_run = last_job_run
        if schedule is not None:
            self.schedule = schedule
        if integration_details is not None:
            self.integration_details = integration_details

    @property
    def id(self):
        """Gets the id of this DecoratedProductionJob.  # noqa: E501

        The id of this job  # noqa: E501

        :return: The id of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedProductionJob.

        The id of this job  # noqa: E501

        :param id: The id of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this DecoratedProductionJob.  # noqa: E501

        Name of the job  # noqa: E501

        :return: The name of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DecoratedProductionJob.

        Name of the job  # noqa: E501

        :param name: The name of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def description(self):
        """Gets the description of this DecoratedProductionJob.  # noqa: E501

        Description of the job  # noqa: E501

        :return: The description of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this DecoratedProductionJob.

        Description of the job  # noqa: E501

        :param description: The description of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedProductionJob.  # noqa: E501

        The time this job was created  # noqa: E501

        :return: The created_at of this DecoratedProductionJob.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedProductionJob.

        The time this job was created  # noqa: E501

        :param created_at: The created_at of this DecoratedProductionJob.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def creator_id(self):
        """Gets the creator_id of this DecoratedProductionJob.  # noqa: E501

        The id of the user who created this job  # noqa: E501

        :return: The creator_id of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this DecoratedProductionJob.

        The id of the user who created this job  # noqa: E501

        :param creator_id: The creator_id of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def config(self):
        """Gets the config of this DecoratedProductionJob.  # noqa: E501

        The config that was used to create this job  # noqa: E501

        :return: The config of this DecoratedProductionJob.  # noqa: E501
        :rtype: ProductionJobConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this DecoratedProductionJob.

        The config that was used to create this job  # noqa: E501

        :param config: The config of this DecoratedProductionJob.  # noqa: E501
        :type: ProductionJobConfig
        """
        if self.local_vars_configuration.client_side_validation and config is None:  # noqa: E501
            raise ValueError("Invalid value for `config`, must not be `None`")  # noqa: E501

        self._config = config

    @property
    def job_queue_config(self):
        """Gets the job_queue_config of this DecoratedProductionJob.  # noqa: E501

        Job Queue configuration of this job (if applicable)  # noqa: E501

        :return: The job_queue_config of this DecoratedProductionJob.  # noqa: E501
        :rtype: JobQueueConfig
        """
        return self._job_queue_config

    @job_queue_config.setter
    def job_queue_config(self, job_queue_config):
        """Sets the job_queue_config of this DecoratedProductionJob.

        Job Queue configuration of this job (if applicable)  # noqa: E501

        :param job_queue_config: The job_queue_config of this DecoratedProductionJob.  # noqa: E501
        :type: JobQueueConfig
        """

        self._job_queue_config = job_queue_config

    @property
    def state(self):
        """Gets the state of this DecoratedProductionJob.  # noqa: E501

        The current state of this job  # noqa: E501

        :return: The state of this DecoratedProductionJob.  # noqa: E501
        :rtype: DecoratedProductionJobStateTransition
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this DecoratedProductionJob.

        The current state of this job  # noqa: E501

        :param state: The state of this DecoratedProductionJob.  # noqa: E501
        :type: DecoratedProductionJobStateTransition
        """
        if self.local_vars_configuration.client_side_validation and state is None:  # noqa: E501
            raise ValueError("Invalid value for `state`, must not be `None`")  # noqa: E501

        self._state = state

    @property
    def project_id(self):
        """Gets the project_id of this DecoratedProductionJob.  # noqa: E501

        Id of the project this job will start clusters in  # noqa: E501

        :return: The project_id of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this DecoratedProductionJob.

        Id of the project this job will start clusters in  # noqa: E501

        :param project_id: The project_id of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def last_job_run_id(self):
        """Gets the last_job_run_id of this DecoratedProductionJob.  # noqa: E501

        The id of the last job run  # noqa: E501

        :return: The last_job_run_id of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._last_job_run_id

    @last_job_run_id.setter
    def last_job_run_id(self, last_job_run_id):
        """Sets the last_job_run_id of this DecoratedProductionJob.

        The id of the last job run  # noqa: E501

        :param last_job_run_id: The last_job_run_id of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """

        self._last_job_run_id = last_job_run_id

    @property
    def schedule_id(self):
        """Gets the schedule_id of this DecoratedProductionJob.  # noqa: E501

        If the job was launched via Scheduled job, this will contain the id of that schedule.  # noqa: E501

        :return: The schedule_id of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._schedule_id

    @schedule_id.setter
    def schedule_id(self, schedule_id):
        """Sets the schedule_id of this DecoratedProductionJob.

        If the job was launched via Scheduled job, this will contain the id of that schedule.  # noqa: E501

        :param schedule_id: The schedule_id of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """

        self._schedule_id = schedule_id

    @property
    def job_queue_id(self):
        """Gets the job_queue_id of this DecoratedProductionJob.  # noqa: E501

        Id of the job queue this job is being enqueued to  # noqa: E501

        :return: The job_queue_id of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._job_queue_id

    @job_queue_id.setter
    def job_queue_id(self, job_queue_id):
        """Sets the job_queue_id of this DecoratedProductionJob.

        Id of the job queue this job is being enqueued to  # noqa: E501

        :param job_queue_id: The job_queue_id of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """

        self._job_queue_id = job_queue_id

    @property
    def overview_url(self):
        """Gets the overview_url of this DecoratedProductionJob.  # noqa: E501

        URL to the job overview page in the Anyscale console  # noqa: E501

        :return: The overview_url of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._overview_url

    @overview_url.setter
    def overview_url(self, overview_url):
        """Sets the overview_url of this DecoratedProductionJob.

        URL to the job overview page in the Anyscale console  # noqa: E501

        :param overview_url: The overview_url of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and overview_url is None:  # noqa: E501
            raise ValueError("Invalid value for `overview_url`, must not be `None`")  # noqa: E501

        self._overview_url = overview_url

    @property
    def is_service(self):
        """Gets the is_service of this DecoratedProductionJob.  # noqa: E501

        Indicates if this job is runs with indefinitely with HA  # noqa: E501

        :return: The is_service of this DecoratedProductionJob.  # noqa: E501
        :rtype: bool
        """
        return self._is_service

    @is_service.setter
    def is_service(self, is_service):
        """Sets the is_service of this DecoratedProductionJob.

        Indicates if this job is runs with indefinitely with HA  # noqa: E501

        :param is_service: The is_service of this DecoratedProductionJob.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and is_service is None:  # noqa: E501
            raise ValueError("Invalid value for `is_service`, must not be `None`")  # noqa: E501

        self._is_service = is_service

    @property
    def url(self):
        """Gets the url of this DecoratedProductionJob.  # noqa: E501

        URL to access deployment running in service  # noqa: E501

        :return: The url of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this DecoratedProductionJob.

        URL to access deployment running in service  # noqa: E501

        :param url: The url of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """

        self._url = url

    @property
    def token(self):
        """Gets the token of this DecoratedProductionJob.  # noqa: E501

        Token used to authenticate user service if it is accessible to public internet. This field will beempty if user service is not publically accessible.  # noqa: E501

        :return: The token of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._token

    @token.setter
    def token(self, token):
        """Sets the token of this DecoratedProductionJob.

        Token used to authenticate user service if it is accessible to public internet. This field will beempty if user service is not publically accessible.  # noqa: E501

        :param token: The token of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """

        self._token = token

    @property
    def access(self):
        """Gets the access of this DecoratedProductionJob.  # noqa: E501

        Whether service can be accessed by public internet traffic.  # noqa: E501

        :return: The access of this DecoratedProductionJob.  # noqa: E501
        :rtype: UserServiceAccessTypes
        """
        return self._access

    @access.setter
    def access(self, access):
        """Sets the access of this DecoratedProductionJob.

        Whether service can be accessed by public internet traffic.  # noqa: E501

        :param access: The access of this DecoratedProductionJob.  # noqa: E501
        :type: UserServiceAccessTypes
        """

        self._access = access

    @property
    def healthcheck_url(self):
        """Gets the healthcheck_url of this DecoratedProductionJob.  # noqa: E501

        The healthcheck url. Anyscale will poll this url to determine whether the service is healthy or not. Only present for services  # noqa: E501

        :return: The healthcheck_url of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._healthcheck_url

    @healthcheck_url.setter
    def healthcheck_url(self, healthcheck_url):
        """Sets the healthcheck_url of this DecoratedProductionJob.

        The healthcheck url. Anyscale will poll this url to determine whether the service is healthy or not. Only present for services  # noqa: E501

        :param healthcheck_url: The healthcheck_url of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """

        self._healthcheck_url = healthcheck_url

    @property
    def archived_at(self):
        """Gets the archived_at of this DecoratedProductionJob.  # noqa: E501

        The time in which this instance is archived.  # noqa: E501

        :return: The archived_at of this DecoratedProductionJob.  # noqa: E501
        :rtype: datetime
        """
        return self._archived_at

    @archived_at.setter
    def archived_at(self, archived_at):
        """Sets the archived_at of this DecoratedProductionJob.

        The time in which this instance is archived.  # noqa: E501

        :param archived_at: The archived_at of this DecoratedProductionJob.  # noqa: E501
        :type: datetime
        """

        self._archived_at = archived_at

    @property
    def cloud_id(self):
        """Gets the cloud_id of this DecoratedProductionJob.  # noqa: E501

        The id of the cloud to which the job belongs.  # noqa: E501

        :return: The cloud_id of this DecoratedProductionJob.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this DecoratedProductionJob.

        The id of the cloud to which the job belongs.  # noqa: E501

        :param cloud_id: The cloud_id of this DecoratedProductionJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def project(self):
        """Gets the project of this DecoratedProductionJob.  # noqa: E501

        The project in which this production job lives  # noqa: E501

        :return: The project of this DecoratedProductionJob.  # noqa: E501
        :rtype: MiniProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this DecoratedProductionJob.

        The project in which this production job lives  # noqa: E501

        :param project: The project of this DecoratedProductionJob.  # noqa: E501
        :type: MiniProject
        """
        if self.local_vars_configuration.client_side_validation and project is None:  # noqa: E501
            raise ValueError("Invalid value for `project`, must not be `None`")  # noqa: E501

        self._project = project

    @property
    def creator(self):
        """Gets the creator of this DecoratedProductionJob.  # noqa: E501

        The creator of this job  # noqa: E501

        :return: The creator of this DecoratedProductionJob.  # noqa: E501
        :rtype: MiniUser
        """
        return self._creator

    @creator.setter
    def creator(self, creator):
        """Sets the creator of this DecoratedProductionJob.

        The creator of this job  # noqa: E501

        :param creator: The creator of this DecoratedProductionJob.  # noqa: E501
        :type: MiniUser
        """
        if self.local_vars_configuration.client_side_validation and creator is None:  # noqa: E501
            raise ValueError("Invalid value for `creator`, must not be `None`")  # noqa: E501

        self._creator = creator

    @property
    def last_job_run(self):
        """Gets the last_job_run of this DecoratedProductionJob.  # noqa: E501

        The last job run that a production job ran with metadata  # noqa: E501

        :return: The last_job_run of this DecoratedProductionJob.  # noqa: E501
        :rtype: MiniJobRun
        """
        return self._last_job_run

    @last_job_run.setter
    def last_job_run(self, last_job_run):
        """Sets the last_job_run of this DecoratedProductionJob.

        The last job run that a production job ran with metadata  # noqa: E501

        :param last_job_run: The last_job_run of this DecoratedProductionJob.  # noqa: E501
        :type: MiniJobRun
        """

        self._last_job_run = last_job_run

    @property
    def schedule(self):
        """Gets the schedule of this DecoratedProductionJob.  # noqa: E501

        The schedule this job was launched by, or None if not launched by a schedule.  # noqa: E501

        :return: The schedule of this DecoratedProductionJob.  # noqa: E501
        :rtype: MiniSchedule
        """
        return self._schedule

    @schedule.setter
    def schedule(self, schedule):
        """Sets the schedule of this DecoratedProductionJob.

        The schedule this job was launched by, or None if not launched by a schedule.  # noqa: E501

        :param schedule: The schedule of this DecoratedProductionJob.  # noqa: E501
        :type: MiniSchedule
        """

        self._schedule = schedule

    @property
    def integration_details(self):
        """Gets the integration_details of this DecoratedProductionJob.  # noqa: E501

        Integration details (eg: W&B run URL) associated with this job.  # noqa: E501

        :return: The integration_details of this DecoratedProductionJob.  # noqa: E501
        :rtype: IntegrationDetails
        """
        return self._integration_details

    @integration_details.setter
    def integration_details(self, integration_details):
        """Sets the integration_details of this DecoratedProductionJob.

        Integration details (eg: W&B run URL) associated with this job.  # noqa: E501

        :param integration_details: The integration_details of this DecoratedProductionJob.  # noqa: E501
        :type: IntegrationDetails
        """

        self._integration_details = integration_details

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedProductionJob):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedProductionJob):
            return True

        return self.to_dict() != other.to_dict()



from click import ClickException

from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.collaborator_type import CollaboratorType


def get_user_id_by_email(api_client: DefaultApi, email: str) -> str:
    users = api_client.list_organization_collaborators_api_v2_organization_collaborators_get(
        email=email, collaborator_type=CollaboratorType.ONLY_USER_ACCOUNTS,
    ).results
    if len(users) == 0:
        raise ClickException(f"No user found with email {email}.")

    if len(users) > 1:
        raise ClickException(
            f"Multiple users found with email {email}. Please provide a unique email."
        )
    return users[0].user_id



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class JobWithReport(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'job_id': 'str',
        'job_name': 'str',
        'job_state': 'HaJobStates',
        'job_report': 'JobReport',
        'created_at': 'datetime',
        'finished_at': 'datetime'
    }

    attribute_map = {
        'job_id': 'job_id',
        'job_name': 'job_name',
        'job_state': 'job_state',
        'job_report': 'job_report',
        'created_at': 'created_at',
        'finished_at': 'finished_at'
    }

    def __init__(self, job_id=None, job_name=None, job_state=None, job_report=None, created_at=None, finished_at=None, local_vars_configuration=None):  # noqa: E501
        """JobWithReport - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._job_id = None
        self._job_name = None
        self._job_state = None
        self._job_report = None
        self._created_at = None
        self._finished_at = None
        self.discriminator = None

        self.job_id = job_id
        self.job_name = job_name
        self.job_state = job_state
        if job_report is not None:
            self.job_report = job_report
        self.created_at = created_at
        if finished_at is not None:
            self.finished_at = finished_at

    @property
    def job_id(self):
        """Gets the job_id of this JobWithReport.  # noqa: E501


        :return: The job_id of this JobWithReport.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this JobWithReport.


        :param job_id: The job_id of this JobWithReport.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and job_id is None:  # noqa: E501
            raise ValueError("Invalid value for `job_id`, must not be `None`")  # noqa: E501

        self._job_id = job_id

    @property
    def job_name(self):
        """Gets the job_name of this JobWithReport.  # noqa: E501


        :return: The job_name of this JobWithReport.  # noqa: E501
        :rtype: str
        """
        return self._job_name

    @job_name.setter
    def job_name(self, job_name):
        """Sets the job_name of this JobWithReport.


        :param job_name: The job_name of this JobWithReport.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and job_name is None:  # noqa: E501
            raise ValueError("Invalid value for `job_name`, must not be `None`")  # noqa: E501

        self._job_name = job_name

    @property
    def job_state(self):
        """Gets the job_state of this JobWithReport.  # noqa: E501


        :return: The job_state of this JobWithReport.  # noqa: E501
        :rtype: HaJobStates
        """
        return self._job_state

    @job_state.setter
    def job_state(self, job_state):
        """Sets the job_state of this JobWithReport.


        :param job_state: The job_state of this JobWithReport.  # noqa: E501
        :type: HaJobStates
        """
        if self.local_vars_configuration.client_side_validation and job_state is None:  # noqa: E501
            raise ValueError("Invalid value for `job_state`, must not be `None`")  # noqa: E501

        self._job_state = job_state

    @property
    def job_report(self):
        """Gets the job_report of this JobWithReport.  # noqa: E501


        :return: The job_report of this JobWithReport.  # noqa: E501
        :rtype: JobReport
        """
        return self._job_report

    @job_report.setter
    def job_report(self, job_report):
        """Sets the job_report of this JobWithReport.


        :param job_report: The job_report of this JobWithReport.  # noqa: E501
        :type: JobReport
        """

        self._job_report = job_report

    @property
    def created_at(self):
        """Gets the created_at of this JobWithReport.  # noqa: E501


        :return: The created_at of this JobWithReport.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this JobWithReport.


        :param created_at: The created_at of this JobWithReport.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def finished_at(self):
        """Gets the finished_at of this JobWithReport.  # noqa: E501


        :return: The finished_at of this JobWithReport.  # noqa: E501
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this JobWithReport.


        :param finished_at: The finished_at of this JobWithReport.  # noqa: E501
        :type: datetime
        """

        self._finished_at = finished_at

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, JobWithReport):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, JobWithReport):
            return True

        return self.to_dict() != other.to_dict()
